import os
from unittest.mock import MagicMock

from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.tests.base_test import TestCase


class BlockExecutorTest(TestCase):
    def setUp(self):
        self.pipeline = MagicMock()
        self.pipeline.uuid = 'pipeline_uuid'
        self.pipeline.repo_config = MagicMock()
        self.block_uuid = 'block_uuid'
        self.execution_partition = 'partition'
        self.logger_manager = MagicMock()
        self.logger = MagicMock()
        self.block = MagicMock(spec=Block)
        self.block.type = BlockType.DBT

        self.pipeline.get_block.return_value = self.block
        self.pipeline.repo_config.retry_config = {'retries': 3, 'delay': 1}
        self.pipeline.repo_config.variables_dir = os.path.join(os.getcwd(), 'mage_data')

        self.logger_manager.logger = self.logger
        self.logger_manager.output_logs_to_destination = MagicMock()

        self.block.uuid = self.block_uuid
        self.block.template_runtime_configuration = None
        self.block.conditional_blocks = []
        self.block.callback_block = None
        self.block.callback_blocks = []

        self.block_executor = BlockExecutor(
            self.pipeline,
            self.block_uuid,
            self.execution_partition,
        )
        self.block_executor.logger_manager = self.logger_manager
        self.block_executor.logger = self.logger
        self.block_executor.block = self.block

    def test_execute(self):
        analyze_outputs = False
        callback_url = 'http://example.com/callback'
        global_vars = {'var1': 'value1'}
        update_status = True

        def on_complete(block_uuid):
            pass

        def on_failure(block_uuid, error):
            pass

        def on_start(block_uuid):
            pass

        input_from_output = {'input': 'output'}
        verify_output = True
        retry_config = {'retries': 5, 'delay': 2}
        runtime_arguments = {'arg1': 'value1'}
        template_runtime_configuration = {'config': 'value'}

        self.block_executor._execute_conditional = MagicMock(return_value=True)
        self.block_executor._execute = MagicMock(return_value={'result': 'success'})
        # self.block.run_tests = MagicMock()
        self.block_executor._execute_callback = MagicMock()

        result = self.block_executor.execute(
            analyze_outputs=analyze_outputs,
            callback_url=callback_url,
            global_vars=global_vars,
            update_status=update_status,
            on_complete=on_complete,
            on_failure=on_failure,
            on_start=on_start,
            input_from_output=input_from_output,
            verify_output=verify_output,
            retry_config=retry_config,
            runtime_arguments=runtime_arguments,
            template_runtime_configuration=template_runtime_configuration,
        )

        self.assertEqual(result, {'result': 'success'})
        self.pipeline.get_block.assert_called_once_with(self.block_uuid, check_template=True)
        self.assertEqual(self.block.template_runtime_configuration, template_runtime_configuration)
        self.block_executor._execute_conditional.assert_called_once_with(
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            global_vars=global_vars,
            logging_tags={
                'block_type': BlockType.DBT,
                'block_uuid': self.block_uuid,
                'pipeline_uuid': self.pipeline.uuid,
            },
            pipeline_run=None,
        )
        self.block_executor._execute.assert_called_once_with(
            analyze_outputs=analyze_outputs,
            block_run_id=None,
            callback_url=callback_url,
            global_vars=global_vars,
            input_from_output=input_from_output,
            logging_tags={
                'block_type': BlockType.DBT,
                'block_uuid': self.block_uuid,
                'pipeline_uuid': self.pipeline.uuid,
            },
            pipeline_run_id=None,
            update_status=update_status,
            verify_output=verify_output,
            runtime_arguments=runtime_arguments,
            template_runtime_configuration=template_runtime_configuration,
            dynamic_block_index=None,
            dynamic_block_uuid=None,
            dynamic_upstream_block_uuids=None,
        )
        # self.block.run_tests.assert_called_once_with(
        #     execution_partition=self.execution_partition,
        #     global_vars=global_vars,
        #     logger=self.logger,
        #     logging_tags={
        #         'block_type': BlockType.DBT,
        #         'block_uuid': self.block_uuid,
        #         'pipeline_uuid': self.pipeline.uuid,
        #     },
        #     update_tests=False,
        #     dynamic_block_uuid=dynamic_block_uuid,
        # )
        self.block_executor._execute_callback.assert_called_with(
            'on_success',
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            global_vars=global_vars,
            logging_tags={
                'block_type': BlockType.DBT,
                'block_uuid': self.block_uuid,
                'pipeline_uuid': self.pipeline.uuid,
            },
            pipeline_run=None,
        )
        self.logger_manager.output_logs_to_destination.assert_called_once()

    def test_execute_conditional(self):
        self.block.conditional_blocks = [MagicMock(), MagicMock()]

        self.block.conditional_blocks[0].execute_conditional.return_value = True
        self.block.conditional_blocks[1].execute_conditional.return_value = False

        result = self.block_executor._execute_conditional(
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            global_vars={},
            logging_tags={},
            pipeline_run=None,
        )

        self.assertFalse(result)
        expected_kwargs = dict(
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            execution_partition=self.execution_partition,
            global_vars={},
            logger=self.logger,
            logging_tags={},
            pipeline_run=None,
        )
        self.block.conditional_blocks[0].execute_conditional.assert_called_once_with(
            self.block,
            **expected_kwargs,
        )
        self.block.conditional_blocks[1].execute_conditional.assert_called_once_with(
            self.block,
            **expected_kwargs,
        )

    def test_execute_conditional_exception(self):
        self.block.conditional_blocks = [MagicMock()]

        self.block.conditional_blocks[0].execute_conditional.side_effect = Exception('Error')

        result = self.block_executor._execute_conditional(
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            global_vars={},
            logging_tags={},
            pipeline_run=None,
        )

        self.assertFalse(result)
        self.block.conditional_blocks[0].execute_conditional.assert_called_once_with(
            self.block,
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            execution_partition=self.execution_partition,
            global_vars={},
            logger=self.logger,
            logging_tags={},
            pipeline_run=None,
        )

    def test_execute_callback(self):
        self.block.callback_blocks = [MagicMock(), MagicMock()]
        self.block.callback_block = MagicMock()

        self.block_executor._execute_callback(
            callback='on_success',
            global_vars={},
            logging_tags={},
            pipeline_run=None,
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
        )

        expected_kwargs = dict(
            dynamic_block_index=None,
            dynamic_upstream_block_uuids=None,
            execution_partition=self.execution_partition,
            global_vars={},
            logger=self.logger,
            logging_tags={},
            parent_block=self.block,
            pipeline_run=None,
        )

        self.block.callback_block.execute_callback.assert_called_once_with(
            'on_success',
            **expected_kwargs,
        )
        self.block.callback_blocks[0].execute_callback.assert_called_once_with(
            'on_success',
            **expected_kwargs,
        )
        self.block.callback_blocks[1].execute_callback.assert_called_once_with(
            'on_success',
            **expected_kwargs,
        )



import asyncio
import base64
import os
import shutil
import subprocess
import uuid
from datetime import datetime
from typing import Any, Dict, List
from urllib.parse import urlparse, urlsplit, urlunsplit

from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.shared.secrets import get_secret_value
from mage_ai.data_preparation.sync import AuthType, GitConfig
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.logger import VerboseFunctionExec

DEFAULT_SSH_KEY_DIRECTORY = os.path.expanduser('~/.ssh')
DEFAULT_KNOWN_HOSTS_FILE = os.path.join(DEFAULT_SSH_KEY_DIRECTORY, 'known_hosts')
REMOTE_NAME = 'mage-repo'

# Git authentication variables
GIT_SSH_PUBLIC_KEY_VAR = 'GIT_SSH_PUBLIC_KEY'
GIT_SSH_PRIVATE_KEY_VAR = 'GIT_SSH_PRIVATE_KEY'
GIT_ACCESS_TOKEN_VAR = 'GIT_ACCESS_TOKEN'


class Git:
    def __init__(self, git_config: GitConfig = None, setup_repo: bool = True) -> None:
        import git

        self.auth_type = AuthType.SSH
        self.git_config = git_config
        self.origin = None
        self.remote_repo_link = None
        self.repo = None
        self.repo_path = os.getcwd()

        if self.git_config:
            self.remote_repo_link = self.git_config.remote_repo_link

            if self.git_config.repo_path:
                self.repo_path = self.git_config.repo_path

            if self.git_config.auth_type:
                self.auth_type = self.git_config.auth_type

        os.makedirs(self.repo_path, exist_ok=True)

        if self.auth_type == AuthType.HTTPS:
            if self.remote_repo_link:
                url = urlsplit(self.remote_repo_link)

            if os.getenv(GIT_ACCESS_TOKEN_VAR):
                token = os.getenv(GIT_ACCESS_TOKEN_VAR)
            elif self.git_config and self.git_config.access_token_secret_name:
                token = get_secret_value(
                    self.git_config.access_token_secret_name,
                    repo_name=get_repo_path(),
                )

            if self.git_config:
                user = self.git_config.username
                url = url._replace(netloc=f'{user}:{token}@{url.netloc}')
                self.remote_repo_link = urlunsplit(url)

        try:
            self.repo = git.Repo(self.repo_path)
        except git.exc.InvalidGitRepositoryError:
            if setup_repo:
                self.__setup_repo()

        if self.repo and self.git_config:
            self.__set_git_config()

        if self.remote_repo_link:
            try:
                self.repo.create_remote(REMOTE_NAME, self.remote_repo_link)
            except git.exc.GitCommandError:
                # if the remote already exists
                pass

            # replace the existing remote url if it is different from the provided url
            self.origin = self.repo.remotes[REMOTE_NAME]
            if self.remote_repo_link not in self.origin.urls:
                self.origin.set_url(self.remote_repo_link)

    @classmethod
    def get_manager(self, user: User = None, setup_repo: bool = True) -> 'Git':
        preferences = get_preferences(user=user)
        git_config = None
        if preferences and preferences.sync_config:
            git_config = GitConfig.load(config=preferences.sync_config)
        return Git(git_config, setup_repo=setup_repo)

    @property
    def current_branch(self) -> Any:
        if not self.repo:
            return None

        return self.repo.git.branch('--show-current')

    @property
    def branches(self) -> List:
        if not self.repo:
            return []

        return [branch.name for branch in self.repo.branches]

    def add_remote(self, name: str, url: str) -> None:
        self.repo.create_remote(name, url)

    def remove_remote(self, name: str) -> None:
        self.repo.git.remote('remove', name)

    async def staged_files(self) -> List[str]:
        if self.repo:
            proc = self.repo.git.diff(
                '--name-only',
                '--cached',
                as_process=True,
            )
            try:
                stdout = await self.__poll_process_with_timeout(
                    proc,
                    error_message='Error fetching untracked files',
                    timeout=10,
                )
                if stdout:
                    return stdout.strip().split('\n')
            except TimeoutError:
                pass
        return []

    async def untracked_files(self, untracked_files: bool = False) -> List[str]:
        if not self.repo:
            return []

        from git.compat import defenc

        # ---------- Modified from GitPython source code -----------
        proc = self.repo.git.status(
            as_process=True,
            porcelain=True,
            untracked_files=untracked_files,
        )
        try:
            stdout = await self.__poll_process_with_timeout(
                proc,
                error_message='Error fetching untracked files',
                timeout=10,
            )
        except TimeoutError:
            lock_file = os.path.join(self.repo_path, '.git', 'index.lock')
            if os.path.exists(lock_file):
                os.remove(lock_file)
            return []
        # Untracked files prefix in porcelain mode
        prefix = '?? '
        files = []
        if stdout:
            for line in stdout.split('\n'):
                # line = line.decode(defenc)
                if not line.startswith(prefix):
                    continue
                filename = line[len(prefix):].rstrip('\n')
                # Special characters are escaped
                if filename[0] == filename[-1] == '"':
                    filename = filename[1:-1]
                    filename = (
                        filename
                        .encode('ascii')
                        .decode('unicode_escape')
                        .encode('latin1')
                        .decode(defenc)
                    )
                files.append(filename)
            # -------------------------------------------------------
            return files

    @property
    def modified_files(self) -> List[str]:
        if not self.repo:
            return []

        return [item.a_path for item in self.repo.index.diff(None)]

    async def check_connection(self) -> None:
        proc = self.repo.git.ls_remote(self.origin.name, as_process=True)

        await self.__poll_process_with_timeout(
            proc,
            error_message='Error connecting to remote, make sure your access token or SSH key is ' +
            'set up properly.',
        )

    def _run_command(self, command: str) -> None:
        proc = subprocess.Popen(args=command, shell=True)
        proc.wait()

    def _remote_command(func) -> None:
        '''
        Decorator method for commands that need to connect to the remote repo. This decorator
        will configure and test SSH settings before executing the Git command.
        '''
        def wrapper(self, *args, **kwargs):
            def add_host_to_known_hosts():
                self.__add_host_to_known_hosts()
                asyncio.run(self.check_connection())

            if self.auth_type == AuthType.SSH:
                url = f'ssh://{self.git_config.remote_repo_link}'
                hostname = urlparse(url).hostname

                private_key_file = self.__create_ssh_keys()
                git_ssh_cmd = f'ssh -i {private_key_file}'
                with self.repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
                    if not os.path.exists(DEFAULT_KNOWN_HOSTS_FILE):
                        self.__add_host_to_known_hosts()
                    try:
                        asyncio.run(self.check_connection())
                    except ChildProcessError as err:
                        if 'Host key verification failed' in str(err):
                            if hostname:
                                add_host_to_known_hosts()
                        else:
                            raise err
                    except TimeoutError:
                        if hostname:
                            add_host_to_known_hosts()
                        else:
                            raise TimeoutError(
                                "Connecting to remote timed out, make sure your SSH key is set up properly"  # noqa: E501
                                " and your repository host is added as a known host. More information here:"  # noqa: E501
                                " https://docs.mage.ai/developing-in-the-cloud/setting-up-git#5-add-github-com-to-known-hosts")  # noqa: E501
                    return func(self, *args, **kwargs)
            else:
                asyncio.run(self.check_connection())
                return func(self, *args, **kwargs)

        return wrapper

    @_remote_command
    def reset(self, branch: str = None) -> None:
        self.origin.fetch()
        if branch is None:
            branch = self.current_branch
        self.repo.git.reset('--hard', f'{self.origin.name}/{branch}')
        self.__pip_install()

    @_remote_command
    def push(self) -> None:
        self.repo.git.push(
            '--set-upstream',
            self.origin.name,
            self.current_branch,
        )

    @_remote_command
    def pull(self) -> None:
        self.origin.pull(self.current_branch)
        self.__pip_install()

    @_remote_command
    def pull_remote_branch(self, remote_name: str, branch_name: str = None):
        import git

        custom_progress = git.remote.RemoteProgress()

        self.set_origin(remote_name)
        remote = self.repo.remotes[remote_name]
        if branch_name and len(branch_name) >= 1:
            try:
                remote.pull(branch_name, custom_progress)
            except git.exc.GitCommandError as err:
                raise err
        else:
            # The following error will occur when no branch name is passed in as the argument.
            # Not sure why, but the pull command still completes.
            # git.exc.GitCommandError: Cmd('git') failed due to: exit code(1)
            # cmdline: git pull -v -- test2
            try:
                remote.pull(progress=custom_progress)
            except git.exc.GitCommandError:
                pass

        return custom_progress

    @_remote_command
    def push_remote_branch(self, remote_name: str, branch_name: str) -> None:
        import git

        custom_progress = git.remote.RemoteProgress()

        self.set_origin(remote_name)
        remote = self.repo.remotes[remote_name]

        remote.push(branch_name, custom_progress)

        return custom_progress

    def set_origin(self, remote_name: str) -> None:
        self.origin = self.repo.remotes[remote_name]

    def status(self) -> str:
        return self.repo.git.status()

    def add_file(self, filename: str, flags: List[str] = None) -> None:
        arr = flags or []
        self.repo.git.add(filename, *arr)

    def checkout_file(self, filename: str) -> None:
        self.repo.git.checkout(filename)

    def diff_file(self, filename: str) -> str:
        return self.repo.git.diff(filename)

    def logs(self, commits: int = 40) -> List[Dict]:
        arr = []

        for idx, commit in enumerate(self.repo.iter_commits()):
            if idx >= commits:
                break

            arr.append(dict(
                author=dict(
                    email=commit.author.email,
                    name=commit.author.name,
                ),
                date=datetime.fromtimestamp(commit.authored_date).isoformat(),
                message=commit.message,
            ))

        return arr

    def delete_branch(self, base_branch_name: str) -> None:
        self.repo.branches[base_branch_name].delete(self.repo, base_branch_name, '-D')

    def merge_branch(self, base_branch_name: str, message: str = None) -> None:
        self.repo.git.merge(self.repo.branches[base_branch_name])
        if message:
            self.repo.index.commit(message)

    def rebase_branch(self, base_branch_name: str, message: str = None) -> None:
        self.repo.git.rebase(self.repo.branches[base_branch_name])
        if message:
            self.repo.index.commit(message)

    def remotes(self, limit: int = 40, user: User = None) -> List[Dict]:
        arr = []

        if not self.repo:
            return arr

        for idx, remote in enumerate(self.repo.remotes):
            from git.exc import GitCommandError

            if idx >= limit:
                break

            error = None
            remote_url = None
            remote_exists = False
            try:
                remote_url = [url for url in remote.urls][0]
                remote_exists = True
            except GitCommandError as err:
                print('WARNING (mage_ai.data_preparation.git.remotes):')
                print(err)

            try:
                remote_refs = remote.refs
                if len(remote_refs) == 0 and user:
                    from mage_ai.data_preparation.git import api

                    access_token = api.get_access_token_for_user(user)
                    if access_token:

                        if remote_exists:
                            token = access_token.token
                            username = api.get_username(token)
                            url = api.build_authenticated_remote_url(
                                remote_url,
                                username,
                                token,
                            )

                            remote.set_url(url)

                            authenticated = False
                            try:
                                api.check_connection(self.repo, url)
                                authenticated = True
                            except Exception as err:
                                print('WARNING (mage_ai.data_preparation.git.remotes):')
                                print(err)

                            if authenticated:
                                try:
                                    remote.fetch()
                                    remote_refs = remote.refs
                                except Exception as err:
                                    print('WARNING (mage_ai.data_preparation.git.remotes):')
                                    print(err)

                refs = []
                for ref in remote_refs:
                    refs.append(dict(
                        name=ref.name,
                        commit=dict(
                            author=dict(
                                email=ref.commit.author.email,
                                name=ref.commit.author.name,
                            ),
                            date=datetime.fromtimestamp(ref.commit.authored_date).isoformat(),
                            message=ref.commit.message,
                        ),
                    ))
            except Exception as err:
                error = err

            try:
                remote.set_url(remote_url)
            except GitCommandError as err:
                print('WARNING (mage_ai.data_preparation.git.remotes):')
                print(err)

            if error:
                raise error

            repository_names = []
            urls = []
            try:
                for url in remote.urls:
                    if url.lower().startswith('https'):
                        repository_names.append('/'.join(url.split('/')[-2:]).replace('.git', ''))
                    urls.append(url)
            except GitCommandError as err:
                print('WARNING (mage_ai.data_preparation.git.remotes):')
                print(err)

            arr.append(dict(
                name=remote.name,
                refs=refs,
                repository_names=repository_names,
                urls=urls,
            ))

        return arr

    def reset_file(self, filename: str) -> None:
        self.repo.git.reset(filename)

    def show_file_from_branch(self, branch: str, filename: str) -> str:
        return self.repo.git.show(f'{branch}:{filename}')

    def commit(self, message, files: List[str] = None) -> None:
        if files:
            for file in files:
                self.repo.git.add(file)
            self.repo.index.commit(message)
        elif self.repo.index.diff(None) or self.repo.untracked_files:
            self.repo.git.add('.')
            self.repo.index.commit(message)

    def commit_message(self, message: str) -> None:
        self.repo.index.commit(message)

    def switch_branch(self, branch) -> None:
        if branch in self.repo.heads:
            self.repo.git.switch(branch)
        else:
            self.repo.git.switch('-c', branch)

    @_remote_command
    def clone(self):
        from git import Repo
        tmp_path = f'{self.repo_path}_{str(uuid.uuid4())}'
        os.mkdir(tmp_path)
        try:
            # Clone remote repo to a tmp folder, and copy over the contents
            # of the tmp folder to the local repo.
            env = {}
            if self.auth_type == AuthType.SSH:
                private_key_file = self.__create_ssh_keys()
                env = {'GIT_SSH_COMMAND': f'ssh -i {private_key_file}'}
            Repo.clone_from(
                self.remote_repo_link,
                to_path=tmp_path,
                origin=REMOTE_NAME,
                env=env,
            )

            shutil.rmtree(os.path.join(self.repo_path, '.git'))
            shutil.copytree(
                tmp_path,
                self.repo_path,
                dirs_exist_ok=True,
                ignore=lambda x, y: ['.preferences.yaml']
            )
            self.repo.git.clean('-fd', exclude='.preferences.yaml')
            self.__pip_install()
        finally:
            shutil.rmtree(tmp_path)

    def __set_git_config(self):
        if self.git_config.username:
            self.repo.config_writer().set_value(
                'user', 'name', self.git_config.username).release()
        if self.git_config.email:
            self.repo.config_writer().set_value(
                'user', 'email', self.git_config.email).release()
        self.repo.config_writer('global').set_value(
            'safe', 'directory', self.repo_path).release()

    def __pip_install(self) -> None:
        requirements_file = os.path.join(
            self.repo.working_dir, 'requirements.txt')

        with VerboseFunctionExec(
            f'Running "pip3 install -r {requirements_file}"',
            verbose=True,
        ):
            try:
                if os.path.exists(requirements_file):
                    cmd = f'pip3 install -r {requirements_file}'
                    self._run_command(cmd)
                print(f'Installing {requirements_file} completed successfully.')
            except Exception as err:
                print(f'Skip installing {requirements_file} due to error: {str(err)}')
                pass

    def __create_ssh_keys(self) -> str:
        if not os.path.exists(DEFAULT_SSH_KEY_DIRECTORY):
            os.mkdir(DEFAULT_SSH_KEY_DIRECTORY, 0o700)
        pubk_secret_name = self.git_config.ssh_public_key_secret_name
        if pubk_secret_name:
            public_key_file = os.path.join(
                DEFAULT_SSH_KEY_DIRECTORY,
                f'id_rsa_{pubk_secret_name}.pub'
            )
            if not os.path.exists(public_key_file):
                try:
                    public_key = get_secret_value(
                        pubk_secret_name,
                        repo_name=get_repo_path(),
                    )
                    if os.getenv(GIT_SSH_PUBLIC_KEY_VAR):
                        public_key = os.getenv(GIT_SSH_PUBLIC_KEY_VAR)
                    if public_key:
                        with open(public_key_file, 'w') as f:
                            f.write(base64.b64decode(public_key).decode('utf-8'))
                        os.chmod(public_key_file, 0o600)
                except Exception:
                    pass
        pk_secret_name = self.git_config.ssh_private_key_secret_name
        private_key_file = os.path.join(DEFAULT_SSH_KEY_DIRECTORY, 'id_rsa')
        if pk_secret_name:
            custom_private_key_file = os.path.join(
                DEFAULT_SSH_KEY_DIRECTORY,
                f'id_rsa_{pk_secret_name}'
            )
            if not os.path.exists(custom_private_key_file):
                try:
                    private_key = get_secret_value(
                        pk_secret_name,
                        repo_name=get_repo_path(),
                    )
                    if os.getenv(GIT_SSH_PRIVATE_KEY_VAR):
                        private_key = os.getenv(GIT_SSH_PRIVATE_KEY_VAR)
                    if private_key:
                        with open(custom_private_key_file, 'w') as f:
                            f.write(base64.b64decode(private_key).decode('utf-8'))
                        os.chmod(custom_private_key_file, 0o600)
                        private_key_file = custom_private_key_file
                except Exception:
                    pass
            else:
                private_key_file = custom_private_key_file

        return private_key_file

    def __setup_repo(self):
        import git
        tmp_path = f'{self.repo_path}_{str(uuid.uuid4())}'
        os.mkdir(tmp_path)
        try:
            # Clone the remote repo and copy over the .git folder
            # to initialize the local repository.
            env = {}
            if self.auth_type == AuthType.SSH:
                private_key_file = self.__create_ssh_keys()
                env = {'GIT_SSH_COMMAND': f'ssh -i {private_key_file}'}
                if not self.__add_host_to_known_hosts():
                    raise Exception('Could not add host to known_hosts')
            repo_git = git.cmd.Git(self.repo_path)
            repo_git.update_environment(**env)
            proc = repo_git.clone(
                self.remote_repo_link,
                tmp_path,
                origin=REMOTE_NAME,
                as_process=True,
            )

            asyncio.run(
                self.__poll_process_with_timeout(
                    proc,
                    error_message='Error cloning repo.',
                    timeout=20,
                )
            )

            git_folder = os.path.join(self.repo_path, '.git')
            tmp_git_folder = os.path.join(tmp_path, '.git')
            shutil.move(tmp_git_folder, git_folder)
            self.repo = git.Repo(path=self.repo_path)
        except Exception:
            self.repo = git.Repo.init(self.repo_path)
            # need to commit something to initialize the repository
            self.commit('Initial commit')
        finally:
            shutil.rmtree(tmp_path)

    def __add_host_to_known_hosts(self):
        url = f'ssh://{self.git_config.remote_repo_link}'
        hostname = urlparse(url).hostname
        if hostname:
            cmd = f'ssh-keyscan -t rsa {hostname} >> {DEFAULT_KNOWN_HOSTS_FILE}'
            self._run_command(cmd)
            return True
        return False

    async def __poll_process_with_timeout(
        self,
        proc: subprocess.Popen,
        error_message: str = None,
        timeout: int = 10,
    ) -> str:
        ct = 0
        return_code = None
        while ct < timeout * 2:
            return_code = proc.poll()
            if return_code is not None:
                proc.kill()
                break
            ct += 1
            await asyncio.sleep(0.5)

        if error_message is None:
            error_message = 'Error running Git process'

        if return_code is not None:
            out, err = proc.communicate()
            if return_code != 0:
                message = (
                    err.decode('UTF-8') if err
                    else error_message
                )
                raise ChildProcessError(message)
            else:
                return out.decode('UTF-8') if out else None

        if return_code is None:
            proc.kill()
            raise TimeoutError



from dataclasses import dataclass
from typing import Dict

from mage_ai.shared.config import BaseConfig

# import traceback

DEFAULT_SERVICE_ACCOUNT_NAME = 'default'
ECS_CONTAINER_METADATA_URI_VAR = 'ECS_CONTAINER_METADATA_URI_V4'


@dataclass
class K8sResourceConfig(BaseConfig):
    cpu: str
    memory: str


@dataclass
class K8sExecutorConfig(BaseConfig):
    container_config: Dict = None
    job_name_prefix: str = None
    resource_limits: Dict = None
    resource_requests: Dict = None
    service_account_name: str = DEFAULT_SERVICE_ACCOUNT_NAME

    @classmethod
    def load(self, config_path: str = None, config: Dict = None):
        executor_config = super().load(config_path=config_path, config=config)
        # if executor_config.resource_limits is not None and \
        #         type(executor_config.resource_limits) is dict:
        #     try:
        #         executor_config.resource_limits = K8sResourceConfig.load(
        #             config=executor_config.resource_limits,
        #         )
        #     except Exception:
        #         traceback.print_exc()
        #         executor_config.resource_limits = None
        # if executor_config.resource_requests is not None and \
        #         type(executor_config.resource_requests) is dict:
        #     try:
        #         executor_config.resource_requests = K8sResourceConfig.load(
        #             config=executor_config.resource_requests,
        #         )
        #     except Exception:
        #         traceback.print_exc()
        #         executor_config.resource_requests = None
        return executor_config



import time
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Dict, List

import pandas as pd
import pyarrow as pa
from deltalake.writer import write_deltalake

from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink


@dataclass
class AzureDataLakeConfig(BaseConfig):
    table_uri: str
    account_name: str
    access_key: str
    mode: str = 'append'
    file_type: str = 'delta'


class AzureDataLakeSink(BaseSink):
    config_class = AzureDataLakeConfig

    def write(self, data: Dict):
        self._print(f'Ingest data {data}, time={time.time()}')
        self.write_buffer(data)

    def batch_write(self, data: List[Dict]):
        if not data:
            self._print('No data')
            return
        self._print(f'Batch ingest {len(data)} records, time={time.time()}. \
                      Sample: {data[0]}')
        self.write_buffer(data)

        self.upload_data_to_adls()

    def upload_data_to_adls(self):
        if not self.buffer:
            return
        self._print(f'Upload {len(self.buffer)} records to DeltaTable.')

        df = pd.DataFrame(self.buffer)

        if self.config.file_type == 'delta':
            curr_time = datetime.now(timezone.utc)
            filename = curr_time.strftime('%Y%m%d-%H%M%S')
            try:
                storage_options = {
                    'azure_storage_account_name': self.config.account_name,
                    'azure_storage_access_key': self.config.access_key}

                data = pa.Table.from_pandas(df)

                write_deltalake(table_or_uri=self.config.table_uri,
                                data=data,
                                name=filename,
                                mode=self.config.mode,
                                storage_options=storage_options)

                self._print(f'Data written to {self.config.table_uri}')
            except Exception as e:
                self.clear_buffer()
                raise Exception(e)
        else:
            self.clear_buffer()
            raise Exception(f'File type {self.config.file_type} \
                              is not supported.')

        self.clear_buffer()



from mage_ai.api.presenters.BasePresenter import BasePresenter


class GitFilePresenter(BasePresenter):
    default_attributes = [
        'content',
        'content_from_base',
        'content_from_compare',
        'error',
        'filename',
        'modified',
    ]



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.OauthPresenter import OauthPresenter


class OauthPolicy(BasePolicy):
    pass


OauthPolicy.allow_actions([
    constants.CREATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())


OauthPolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
    OauthScope.CLIENT_PUBLIC,
], condition=lambda policy: policy.has_at_least_viewer_role())


OauthPolicy.allow_read(OauthPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_viewer_role())


OauthPolicy.allow_read(OauthPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
    OauthScope.CLIENT_PUBLIC,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())


OauthPolicy.allow_write([
    'provider',
    'token',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_viewer_role())


OauthPolicy.allow_query([
    'redirect_uri',
], on_action=[
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
    OauthScope.CLIENT_PUBLIC,
], condition=lambda policy: policy.has_at_least_viewer_role())



from typing import Dict

from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.services.azure.container_instance import container_instance
from mage_ai.shared.hash import merge_dict


class AzureContainerInstanceExecutor(BlockExecutor):
    RETRYABLE = False

    def __init__(self, pipeline, block_uuid: str, execution_partition: str = None):
        super().__init__(pipeline, block_uuid, execution_partition=execution_partition)
        self.executor_config = self.pipeline.repo_config.azure_container_instance_config or dict()
        if self.block.executor_config is not None:
            self.executor_config = merge_dict(self.executor_config, self.block.executor_config)

    def execute(
        self,
        block_run_id: int = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> None:
        cmd = self._run_commands(
            block_run_id=block_run_id,
            global_vars=global_vars,
            **kwargs,
        )
        container_instance.run_job(
            ' '.join(cmd),
            f'mage-data-prep-block-{block_run_id}',
            self.executor_config,
        )



"""Update unique constraint for secret

Revision ID: 90d978a8aef8
Revises: dfe49d040487
Create Date: 2023-07-19 09:54:52.174582

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine.reflection import Inspector


# revision identifiers, used by Alembic.
revision = '90d978a8aef8'
down_revision = 'dfe49d040487'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        bind = op.get_bind()
        insp = Inspector.from_engine(bind.engine)
        with op.batch_alter_table('secret', schema=None) as batch_op:
            unique_constraints = insp.get_unique_constraints('secret')
            for constraint in unique_constraints:
                batch_op.drop_constraint(constraint['name'], type_='unique')
    except ValueError:
        with op.batch_alter_table('secret', schema=None) as batch_op:
            batch_op.add_column(sa.Column('name_new', sa.String(length=255), nullable=True))
        with op.batch_alter_table('secret', schema=None) as batch_op:
            batch_op.execute('UPDATE secret SET name_new = name')
        with op.batch_alter_table('secret', schema=None) as batch_op:
            batch_op.drop_column('name')
            batch_op.add_column(sa.Column('name', sa.String(length=255), nullable=True))
        with op.batch_alter_table('secret', schema=None) as batch_op:
            batch_op.execute('UPDATE secret SET name = name_new')
        with op.batch_alter_table('secret', schema=None) as batch_op:
            batch_op.drop_column('name_new')
    
    with op.batch_alter_table('secret', schema=None) as batch_op:
        batch_op.create_unique_constraint('name_unique', columns=['name'])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from mage_ai.cache.constants import (
    MAGE_CACHE_DIRECTORY_DEFAULT,
    MAGE_CACHE_DIRECTORY_ENVIRONMENT_VARIABLE_NAME,
)
from mage_ai.data_preparation.repo_manager import RepoConfig, get_repo_config
from mage_ai.data_preparation.storage.local_storage import LocalStorage
from mage_ai.settings.repo import get_repo_path
from typing import Any, Dict, List, Union
import os


class BaseCache():
    cache_key = None

    def __init__(self, repo_path: str = None, repo_config=None):
        self.repo_path = repo_path or get_repo_path()

        if repo_config is None:
            self.repo_config = get_repo_config(repo_path=self.repo_path)
        elif type(repo_config) is dict:
            self.repo_config = RepoConfig.from_dict(repo_config)
        else:
            self.repo_config = repo_config

        self._storage = None

    @property
    def storage(self) -> LocalStorage:
        if not self._storage:
            self._storage = LocalStorage()

        return self._storage

    def exists(self) -> bool:
        return self.get(self.cache_key) is not None

    def get(self, key: str) -> Union[Dict, List]:
        value_fetched = self.storage.read_json_file(self.__build_path(key), None)
        return value_fetched

    def load_all_data(self):
        return self.get(self.cache_key)

    def set(self, key: str, value: Any) -> None:
        self.storage.write_json_file(self.__build_path(key), value)

    def get_set(self, key: str, value: Any) -> Union[Dict, List]:
        value_fetched = self.get(key)

        if value_fetched is None:
            self.set(key, value)
            value_fetched = value

        return value_fetched

    def __build_path(self, key: str) -> str:
        dir_path = os.getenv(MAGE_CACHE_DIRECTORY_ENVIRONMENT_VARIABLE_NAME) or os.path.join(
            self.repo_config.variables_dir,
            MAGE_CACHE_DIRECTORY_DEFAULT,
        )

        return os.path.join(dir_path, key)



import os
import sys

from mage_ai.shared.environments import is_test

"""
Moved from repo_manager because repo_manager has too many dependencies
which can cause circular import errors.
"""

if is_test():
    DEFAULT_MAGE_DATA_DIR = './'
else:
    DEFAULT_MAGE_DATA_DIR = '~/.mage_data'
MAGE_DATA_DIR_ENV_VAR = 'MAGE_DATA_DIR'
REPO_PATH_ENV_VAR = 'MAGE_REPO_PATH'


def get_repo_path() -> str:
    return os.getenv(REPO_PATH_ENV_VAR) or os.getcwd()


def set_repo_path(repo_path: str) -> None:
    os.environ[REPO_PATH_ENV_VAR] = repo_path
    sys.path.append(os.path.dirname(repo_path))


def get_repo_name() -> str:
    return os.path.basename(get_repo_path())


def get_data_dir() -> str:
    return os.getenv(MAGE_DATA_DIR_ENV_VAR) or DEFAULT_MAGE_DATA_DIR



import os
from typing import Dict, List

import yaml
from kubernetes import client, config

from mage_ai.cluster_manager.constants import (
    CLOUD_SQL_CONNECTION_NAME,
    CONNECTION_URL_SECRETS_NAME,
    DB_SECRETS_NAME,
    GCP_BACKEND_CONFIG_ANNOTATION,
    KUBE_NAMESPACE,
    KUBE_SERVICE_GCP_BACKEND_CONFIG,
    KUBE_SERVICE_TYPE,
    NODE_PORT_SERVICE_TYPE,
    SERVICE_ACCOUNT_CREDENTIAL_FILE_PATH,
    SERVICE_ACCOUNT_SECRETS_NAME,
)
from mage_ai.data_preparation.repo_manager import ProjectType
from mage_ai.orchestration.constants import (
    DATABASE_CONNECTION_URL_ENV_VAR,
    DB_NAME,
    DB_PASS,
    DB_USER,
)
from mage_ai.services.k8s.constants import (
    DEFAULT_NAMESPACE,
    DEFAULT_SERVICE_ACCOUNT_NAME,
    DEFAULT_STORAGE_CLASS_NAME,
    KUBE_POD_NAME_ENV_VAR,
)
from mage_ai.settings import MAGE_SETTINGS_ENVIRONMENT_VARIABLES
from mage_ai.shared.array import find


class WorkloadManager:
    def __init__(self, namespace: str = DEFAULT_NAMESPACE):
        self.load_config()
        self.core_client = client.CoreV1Api()
        self.apps_client = client.AppsV1Api()

        self.namespace = namespace
        if not self.namespace:
            self.namespace = DEFAULT_NAMESPACE

        try:
            self.pod_config = self.core_client.read_namespaced_pod(
                name=os.getenv(KUBE_POD_NAME_ENV_VAR),
                namespace=self.namespace,
            )
        except Exception:
            self.pod_config = None

    @classmethod
    def load_config(cls) -> bool:
        try:
            config.load_incluster_config()
            return True
        except Exception:
            pass

        try:
            config.load_kube_config()
        except Exception:
            pass

        return False

    def list_workloads(self):
        services = self.core_client.list_namespaced_service(self.namespace).items
        workloads_list = []

        pods = self.core_client.list_namespaced_pod(self.namespace).items
        pod_mapping = dict()
        for pod in pods:
            try:
                name = pod.metadata.labels.get('app')
                pod_mapping[name] = pod
            except Exception:
                pass
        for service in services:
            try:
                labels = service.metadata.labels
                if not labels.get('dev-instance'):
                    continue
                name = labels.get('app')
                service_type = service.spec.type
                workload = dict(
                    name=name,
                    type=service_type,
                )
                pod = pod_mapping.get(name)
                if pod:
                    status = pod.status.phase
                    workload['status'] = status.upper()

                    node_name = pod.spec.node_name
                    ip = None
                    if service_type == 'NodePort':
                        try:
                            if node_name:
                                items = self.core_client.list_node(
                                    field_selector=f'metadata.name={node_name}').items
                                node = items[0]
                                ip = find(
                                    lambda a: a.type == 'ExternalIP',
                                    node.status.addresses
                                ).address
                                if ip:
                                    node_port = service.spec.ports[0].node_port
                                    workload['ip'] = f'{ip}:{node_port}'
                        except Exception:
                            pass
                else:
                    workload['status'] = 'UNAVAILABLE'

                workloads_list.append(workload)
            except Exception:
                pass

        return workloads_list

    def create_workload(
        self,
        name: str,
        project_type: str = ProjectType.STANDALONE,
        project_uuid: str = None,
        **kwargs,
    ):
        container_config_yaml = kwargs.get('container_config')
        container_config = dict()
        if container_config_yaml:
            container_config = yaml.full_load(container_config_yaml)

        parameters = self.__get_configurable_parameters(**kwargs)
        service_account_name = parameters.get(
            'service_account_name',
            DEFAULT_SERVICE_ACCOUNT_NAME,
        )
        storage_class_name = parameters.get(
            'storage_class_name',
            DEFAULT_STORAGE_CLASS_NAME,
        )
        storage_access_mode = parameters.get('storage_access_mode', 'ReadWriteOnce')
        storage_request_size = parameters.get('storage_request_size', '2Gi')

        env_vars = self.__populate_env_vars(
            name,
            project_type=project_type,
            project_uuid=project_uuid,
            container_config=container_config,
        )
        container_config['env'] = env_vars

        containers = [
            {
                'name': f'{name}-container',
                'image': 'mageai/mageai:latest',
                'ports': [
                    {
                        'containerPort': 6789,
                        'name': 'web'
                    }
                ],
                'volumeMounts': [
                    {
                        'name': 'mage-data',
                        'mountPath': '/home/src'
                    }
                ],
                **container_config,
            }
        ]

        volumes = []
        if os.getenv(SERVICE_ACCOUNT_SECRETS_NAME):
            credential_file_path = os.getenv(
                SERVICE_ACCOUNT_CREDENTIAL_FILE_PATH,
                'service_account.json',
            )
            containers.append(
                {
                    'name': 'cloud-sql-proxy',
                    'image': 'gcr.io/cloudsql-docker/gce-proxy:1.28.0',
                    'command': [
                        '/cloud_sql_proxy',
                        '-log_debug_stdout',
                        f'-instances={os.getenv(CLOUD_SQL_CONNECTION_NAME)}=tcp:5432',
                        f'-credential_file=/secrets/{credential_file_path}'
                    ],
                    'securityContext': {
                        'runAsNonRoot': True
                    },
                    'resources': {
                        'requests': {
                            'memory': '1Gi',
                            'cpu': '1'
                        }
                    },
                    'volumeMounts': [
                        {
                            'name': 'service-account-volume',
                            'mountPath': '/secrets/',
                            'readOnly': True
                        }
                    ]
                }
            )
            volumes.append(
                {
                    'name': 'service-account-volume',
                    'secret': {
                        'secretName': os.getenv(SERVICE_ACCOUNT_SECRETS_NAME)
                    }
                }
            )

        pod_spec = self.pod_config.spec.to_dict() if self.pod_config else dict()
        stateful_set_template_spec = dict(
            imagePullSecrets=pod_spec.get('image_pull_secrets'),
            terminationGracePeriodSeconds=10,
            containers=containers,
            volumes=volumes,
        )
        if service_account_name:
            stateful_set_template_spec['serviceAccountName'] = service_account_name

        stateful_set = {
            'apiVersion': 'apps/v1',
            'kind': 'StatefulSet',
            'metadata': {
                'name': name,
                'labels': {
                    'app': name
                }
            },
            'spec': {
                'selector': {
                    'matchLabels': {
                        'app': name
                    }
                },
                'replicas': 1,
                'minReadySeconds': 10,
                'template': {
                    'metadata': {
                        'labels': {
                            'app': name
                        }
                    },
                    'spec': stateful_set_template_spec
                },
                'volumeClaimTemplates': [
                    {
                        'metadata': {
                            'name': 'mage-data'
                        },
                        'spec': {
                            'accessModes': [storage_access_mode],
                            'storageClassName': storage_class_name,
                            'resources': {
                                'requests': {
                                    'storage': storage_request_size
                                }
                            }
                        }
                    }
                ]
            }
        }

        self.apps_client.create_namespaced_stateful_set(self.namespace, stateful_set)

        service_name = f'{name}-service'

        annotations = {}
        if os.getenv(KUBE_SERVICE_GCP_BACKEND_CONFIG):
            annotations[GCP_BACKEND_CONFIG_ANNOTATION] = \
                os.getenv(KUBE_SERVICE_GCP_BACKEND_CONFIG)

        service = {
            'apiVersion': 'v1',
            'kind': 'Service',
            'metadata': {
                'name': service_name,
                'labels': {
                    'app': name,
                    'dev-instance': '1',
                },
                'annotations': annotations
            },
            'spec': {
                'ports': [
                    {
                        'protocol': 'TCP',
                        'port': 6789,
                    }
                ],
                'selector': {
                    'app': name
                },
                'type': os.getenv(KUBE_SERVICE_TYPE, NODE_PORT_SERVICE_TYPE)
            }
        }

        return self.core_client.create_namespaced_service(self.namespace, service)

    def delete_workload(self, name: str):
        self.apps_client.delete_namespaced_stateful_set(name, self.namespace)
        self.core_client.delete_namespaced_service(f'{name}-service', self.namespace)

    def __populate_env_vars(
        self,
        name,
        project_type: str = 'standalone',
        project_uuid: str = None,
        container_config: Dict = None
    ) -> List:
        env_vars = [
            {
                'name': 'USER_CODE_PATH',
                'value': name,
            }
        ]
        if project_type:
            env_vars.append({
                'name': 'PROJECT_TYPE',
                'value': project_type,
            })
        if project_uuid:
            env_vars.append({
                'name': 'PROJECT_UUID',
                'value': project_uuid,
            })

        connection_url_secrets_name = os.getenv(CONNECTION_URL_SECRETS_NAME)
        if connection_url_secrets_name:
            env_vars.append(
                {
                    'name': DATABASE_CONNECTION_URL_ENV_VAR,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': connection_url_secrets_name,
                            'key': 'connection_url'
                        }
                    }
                }
            )

        for var in MAGE_SETTINGS_ENVIRONMENT_VARIABLES + [
            DATABASE_CONNECTION_URL_ENV_VAR,
            KUBE_NAMESPACE,
        ]:
            if os.getenv(var) is not None:
                env_vars.append({
                    'name': var,
                    'value': str(os.getenv(var)),
                })

        # For connecting to CloudSQL PostgreSQL database.
        db_secrets_name = os.getenv(DB_SECRETS_NAME)
        if db_secrets_name:
            env_vars.extend([
                {
                    'name': DB_USER,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': db_secrets_name,
                            'key': 'username'
                        }
                    }
                },
                {
                    'name': DB_PASS,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': db_secrets_name,
                            'key': 'password'
                        }
                    }
                },
                {
                    'name': DB_NAME,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': db_secrets_name,
                            'key': 'database'
                        }
                    }
                }
            ])

        if container_config and 'env' in container_config:
            env_vars += container_config['env']

        return env_vars

    def __get_configurable_parameters(self, **kwargs) -> Dict:
        service_account_name_default = None
        storage_class_name_default = None
        storage_access_mode_default = None
        storage_request_size_default = None
        try:
            service_account_name_default = self.pod_config.spec.service_account_name

            pvc_name = find(
                lambda v: v.persistent_volume_claim is not None,
                self.pod_config.spec.volumes,
            ).persistent_volume_claim.claim_name

            pvc = self.core_client.read_namespaced_persistent_volume_claim(
                name=pvc_name,
                namespace=self.namespace,
            )

            storage_class_name_default = pvc.spec.storage_class_name
            storage_access_mode_default = pvc.spec.access_modes[0]
            storage_request_size_default = pvc.spec.resources.requests.get('storage')
        except Exception:
            pass

        storage_request_size = kwargs.get('storage_request_size')
        if storage_request_size is None:
            storage_request_size = storage_request_size_default
        else:
            storage_request_size = f'{storage_request_size}Gi'

        return dict(
            service_account_name=kwargs.get('service_account_name', service_account_name_default),
            storage_class_name=kwargs.get('storage_class_name', storage_class_name_default),
            storage_access_mode=kwargs.get('storage_access_mode', storage_access_mode_default),
            storage_request_size=storage_request_size,
        )

    def __create_persistent_volume(
        self,
        name,
        volume_host_path=None,
        storage_request_size='2Gi',
        access_mode=None,
    ):
        nodes = self.core_client.list_node().items
        hostnames = [node.metadata.labels['kubernetes.io/hostname'] for node in nodes]
        access_modes = ['ReadWriteOnce'] if access_mode is None else [access_mode]
        pv = {
            'apiVersion': 'v1',
            'kind': 'PersistentVolume',
            'metadata': {
                'name': f'{name}-pv'
            },
            'spec': {
                'capacity': {
                    'storage': storage_request_size,
                },
                'volumeMode': 'Filesystem',
                'accessModes': access_modes,
                'persistentVolumeReclaimPolicy': 'Delete',
                'storageClassName': f'{name}-local-storage',
                'local': {
                    'path': volume_host_path,
                },
                'nodeAffinity': {
                    'required': {
                        'nodeSelectorTerms': [
                            {
                                'matchExpressions': [
                                    {
                                        'key': 'kubernetes.io/hostname',
                                        'operator': 'In',
                                        'values': hostnames
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }
        self.core_client.create_persistent_volume(pv)



from typing import Dict

import requests


def get_user_info(token: str) -> Dict:
    url = 'https://graph.microsoft.com/v1.0/me'

    headers = {
        'Content-Type': 'application\\json',
        'Authorization': 'Bearer {}'.format(token)
    }

    resp = requests.get(url, headers=headers)
    result = resp.json()

    return result



import os
import traceback
from datetime import datetime, timedelta
from typing import Any, Dict, List, Set, Tuple

import pytz
from dateutil.relativedelta import relativedelta
from sqlalchemy import desc, func

from mage_ai.data_integrations.utils.scheduler import (
    clear_source_output_files,
    get_extra_variables,
    initialize_state_and_runs,
)
from mage_ai.data_preparation.executors.executor_factory import ExecutorFactory
from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.logging.logger_manager_factory import LoggerManagerFactory
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.utils import (
    create_block_runs_from_dynamic_block,
    is_dynamic_block,
)
from mage_ai.data_preparation.models.constants import ExecutorType, PipelineType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import (
    ScheduleInterval,
    ScheduleStatus,
    ScheduleType,
    get_triggers_by_pipeline,
)
from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.data_preparation.sync.git_sync import get_sync_config
from mage_ai.orchestration.concurrency import ConcurrencyConfig
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.orchestration.db.models.schedules import (
    Backfill,
    BlockRun,
    EventMatcher,
    PipelineRun,
    PipelineSchedule,
)
from mage_ai.orchestration.job_manager import JobType, job_manager
from mage_ai.orchestration.metrics.pipeline_run import calculate_metrics
from mage_ai.orchestration.notification.config import NotificationConfig
from mage_ai.orchestration.notification.sender import NotificationSender
from mage_ai.orchestration.utils.distributed_lock import DistributedLock
from mage_ai.orchestration.utils.git import log_git_sync, run_git_sync
from mage_ai.orchestration.utils.resources import get_compute, get_memory
from mage_ai.server.logger import Logger
from mage_ai.settings import HOSTNAME
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.array import find
from mage_ai.shared.dates import compare
from mage_ai.shared.environments import get_env
from mage_ai.shared.hash import index_by, merge_dict
from mage_ai.shared.retry import retry

MEMORY_USAGE_MAXIMUM = 0.95

lock = DistributedLock()
logger = Logger().new_server_logger(__name__)


class PipelineScheduler:
    def __init__(
        self,
        pipeline_run: PipelineRun,
    ) -> None:
        self.pipeline_run = pipeline_run
        self.pipeline_schedule = pipeline_run.pipeline_schedule
        self.pipeline = Pipeline.get(pipeline_run.pipeline_uuid)

        # Get the list of integration stream if the pipeline is data integration pipeline
        self.streams = []
        if self.pipeline.type == PipelineType.INTEGRATION:
            self.streams = self.pipeline.streams(
                self.pipeline_run.get_variables(
                    extra_variables=get_extra_variables(self.pipeline)
                )
            )

        # Initialize the logger
        self.logger_manager = LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline.uuid,
            partition=self.pipeline_run.execution_partition,
            repo_config=self.pipeline.repo_config,
        )
        self.logger = DictLogger(self.logger_manager.logger)

        # Initialize the notification sender
        self.notification_sender = NotificationSender(
            NotificationConfig.load(
                config=merge_dict(
                    self.pipeline.repo_config.notification_config,
                    self.pipeline.notification_config,
                )
            )
        )

        self.concurrency_config = ConcurrencyConfig.load(
            config=self.pipeline.concurrency_config
        )

        # Other pipeline scheduling settings
        self.allow_blocks_to_fail = (
            self.pipeline_schedule.get_settings().allow_blocks_to_fail
            if self.pipeline_schedule else False
        )

    def start(self, should_schedule: bool = True) -> bool:
        """Start the pipeline run.

        This method starts the pipeline run by performing necessary actions
        * Update the pipeline run status
        * Optionally scheduling the pipeline execution

        Args:
            should_schedule (bool, optional): Flag indicating whether to schedule
                the pipeline execution. Defaults to True.

        Returns:
            bool: Whether the pipeline run is started successfully
        """
        if self.pipeline_run.status == PipelineRun.PipelineRunStatus.RUNNING:
            return True

        tags = self.build_tags()

        is_integration = PipelineType.INTEGRATION == self.pipeline.type

        try:
            block_runs = BlockRun.query.filter(
                BlockRun.pipeline_run_id == self.pipeline_run.id).all()

            if len(block_runs) == 0:
                if is_integration:
                    clear_source_output_files(
                        self.pipeline_run,
                        self.logger,
                    )
                    initialize_state_and_runs(
                        self.pipeline_run,
                        self.logger,
                        self.pipeline_run.get_variables(),
                    )
                else:
                    self.pipeline_run.create_block_runs()
        except Exception as e:
            self.logger.exception(
                'Fail to initialize block runs',
                **merge_dict(tags, dict(
                    error=e,
                )),
            )
            self.pipeline_run.update(status=PipelineRun.PipelineRunStatus.FAILED)
            return False

        self.pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        if should_schedule:
            self.schedule()
        return True

    def stop(self) -> None:
        stop_pipeline_run(
            self.pipeline_run,
            self.pipeline,
        )

    def schedule(self, block_runs: List[BlockRun] = None) -> None:
        if not lock.try_acquire_lock(f'pipeline_run_{self.pipeline_run.id}', timeout=10):
            return

        self.__run_heartbeat()

        for b in self.pipeline_run.block_runs:
            b.refresh()

        if PipelineType.STREAMING == self.pipeline.type:
            self.__schedule_pipeline()
        else:
            if self.pipeline_run.all_blocks_completed(self.allow_blocks_to_fail):
                if PipelineType.INTEGRATION == self.pipeline.type:
                    tags = self.build_tags()
                    calculate_metrics(self.pipeline_run, logger=self.logger, logging_tags=tags)

                if self.pipeline_run.any_blocks_failed():
                    self.pipeline_run.update(
                        status=PipelineRun.PipelineRunStatus.FAILED,
                        completed_at=datetime.now(),
                    )
                    self.notification_sender.send_pipeline_run_failure_message(
                        pipeline=self.pipeline,
                        pipeline_run=self.pipeline_run,
                    )
                else:
                    self.pipeline_run.complete()
                    self.notification_sender.send_pipeline_run_success_message(
                        pipeline=self.pipeline,
                        pipeline_run=self.pipeline_run,
                    )

                self.logger_manager.output_logs_to_destination()

                schedule = PipelineSchedule.get(
                    self.pipeline_run.pipeline_schedule_id,
                )

                if schedule:
                    backfills = schedule.backfills
                    # When all pipeline runs that are associated with backfill is done
                    if len(backfills) >= 1:
                        backfill = backfills[0]
                        if all([PipelineRun.PipelineRunStatus.COMPLETED == pr.status
                                for pr in backfill.pipeline_runs]):
                            backfill.update(
                                completed_at=datetime.now(),
                                status=Backfill.Status.COMPLETED,
                            )
                            schedule.update(
                                status=ScheduleStatus.INACTIVE,
                            )
                    # If running once, update the schedule to inactive when pipeline run is done
                    elif schedule.status == ScheduleStatus.ACTIVE and \
                            schedule.schedule_type == ScheduleType.TIME and \
                            schedule.schedule_interval == ScheduleInterval.ONCE:

                        schedule.update(status=ScheduleStatus.INACTIVE)
            elif self.pipeline_run.any_blocks_failed() and \
                    not self.allow_blocks_to_fail:
                self.pipeline_run.update(
                    status=PipelineRun.PipelineRunStatus.FAILED)
                self.notification_sender.send_pipeline_run_failure_message(
                    pipeline=self.pipeline,
                    pipeline_run=self.pipeline_run,
                )
                # Cancel block runs that are still in progress for the pipeline run.
                cancel_block_runs_and_jobs(self.pipeline_run, self.pipeline)
            elif PipelineType.INTEGRATION == self.pipeline.type:
                self.__schedule_integration_streams(block_runs)
            else:
                self.__schedule_blocks(block_runs)

    def on_block_complete(self, block_uuid: str) -> None:
        block_run = BlockRun.get(pipeline_run_id=self.pipeline_run.id, block_uuid=block_uuid)

        @retry(retries=2, delay=5)
        def update_status():
            block_run.update(
                status=BlockRun.BlockRunStatus.COMPLETED,
                completed_at=datetime.now(),
            )

        update_status()

        self.logger.info(
            f'BlockRun {block_run.id} (block_uuid: {block_uuid}) completes.',
            **self.build_tags(
                block_run_id=block_run.id,
                block_uuid=block_run.block_uuid,
            ),
        )

        self.pipeline_run.refresh()
        if self.pipeline_run.status != PipelineRun.PipelineRunStatus.RUNNING:
            return
        else:
            self.schedule()

    def on_block_complete_without_schedule(self, block_uuid: str) -> None:
        block = self.pipeline.get_block(block_uuid)
        if block and is_dynamic_block(block):
            create_block_runs_from_dynamic_block(
                block,
                self.pipeline_run,
                block_uuid=block.uuid if block.replicated_block else block_uuid,
            )

        block_run = BlockRun.get(pipeline_run_id=self.pipeline_run.id, block_uuid=block_uuid)

        @retry(retries=2, delay=5)
        def update_status():
            block_run.update(
                status=BlockRun.BlockRunStatus.COMPLETED,
                completed_at=datetime.now(),
            )

        update_status()

        self.logger.info(
            f'BlockRun {block_run.id} (block_uuid: {block_uuid}) completes.',
            **self.build_tags(
                block_run_id=block_run.id,
                block_uuid=block_run.block_uuid,
            ),
        )

    def on_block_failure(self, block_uuid: str, **kwargs) -> None:
        block_run = BlockRun.get(pipeline_run_id=self.pipeline_run.id, block_uuid=block_uuid)
        metrics = block_run.metrics or {}

        @retry(retries=2, delay=5)
        def update_status():
            block_run.update(
                metrics=metrics,
                status=BlockRun.BlockRunStatus.FAILED,
            )

        update_status()

        if 'error' in kwargs:
            metrics['error'] = kwargs['error']

        tags = self.build_tags(
            block_run_id=block_run.id,
            block_uuid=block_run.block_uuid,
        )

        self.logger.info(
            f'BlockRun {block_run.id} (block_uuid: {block_uuid}) failed.',
            **tags,
        )

        if not self.allow_blocks_to_fail:
            if PipelineType.INTEGRATION == self.pipeline.type:
                # If a block/stream fails, stop all other streams
                job_manager.kill_pipeline_run_job(self.pipeline_run.id)
                for stream in self.streams:
                    job_manager.kill_integration_stream_job(
                        self.pipeline_run.id,
                        stream.get('tap_stream_id')
                    )

                calculate_metrics(self.pipeline_run, logger=self.logger, logging_tags=tags)

    def memory_usage_failure(self, tags: Dict = None) -> None:
        if tags is None:
            tags = dict()
        msg = 'Memory usage across all pipeline runs has reached or exceeded the maximum '\
            f'limit of {int(MEMORY_USAGE_MAXIMUM * 100)}%.'
        self.logger.info(msg, tags=tags)

        self.stop()

        self.notification_sender.send_pipeline_run_failure_message(
            pipeline=self.pipeline,
            pipeline_run=self.pipeline_run,
            summary=msg,
        )

        if PipelineType.INTEGRATION == self.pipeline.type:
            calculate_metrics(self.pipeline_run, logger=self.logger, logging_tags=tags)

    @property
    def executable_block_runs(self) -> List[BlockRun]:
        """Get the list of executable block runs.

        This property returns a list of block runs that are considered executable for the
        pipeline run. It first builds the block UUIDs for the completed block runs and the block
        runs in progress.
        The method iterates over the initial block runs and determines their executability based
        on the status of their dynamic upstream block runs or upstream blocks.
        * If the block run has dynamic upstream block UUIDs, it checks if all of them are present in
        the finished block UUIDs or completed block UUIDs depending on the `allow_blocks_to_fail`
        flag.
        * If the block run does not have dynamic upstream block UUIDs, it checks if all upstream
        blocks are completed.
        If a block run's upstream blocks all completed, it is added to the list of executable block
        runs.

        Returns:
            List[BlockRun]: A list of executable block runs for the pipeline run.
        """
        def _build_block_uuids(block_runs: List[BlockRun]) -> List[str]:
            arr = set()

            for block_run in block_runs:
                if block_run.status in [
                    BlockRun.BlockRunStatus.COMPLETED,
                    BlockRun.BlockRunStatus.UPSTREAM_FAILED,
                    BlockRun.BlockRunStatus.FAILED,
                ]:
                    block_uuid = block_run.block_uuid
                    block = self.pipeline.get_block(block_uuid)
                    arr.add(block_uuid)

                    # Block runs for replicated blocks have the following block UUID convention:
                    # [block.uuid]:[block.replicated_block]
                    if block and block.replicated_block:
                        arr.add(block.uuid)

            return arr

        completed_block_uuids = _build_block_uuids(self.pipeline_run.completed_block_runs)
        finished_block_uuids = _build_block_uuids(self.pipeline_run.block_runs)

        executable_block_runs = list()
        for block_run in self.pipeline_run.initial_block_runs:
            completed = False

            dynamic_upstream_block_uuids = block_run.metrics and block_run.metrics.get(
                'dynamic_upstream_block_uuids',
            )

            if dynamic_upstream_block_uuids:
                if self.allow_blocks_to_fail:
                    completed = all(uuid in finished_block_uuids
                                    for uuid in dynamic_upstream_block_uuids)
                else:
                    completed = all(uuid in completed_block_uuids
                                    for uuid in dynamic_upstream_block_uuids)
            else:
                block = self.pipeline.get_block(block_run.block_uuid)
                completed = block is not None and \
                    block.all_upstream_blocks_completed(completed_block_uuids)

            if completed:
                executable_block_runs.append(block_run)

        return executable_block_runs

    def build_tags(self, **kwargs):
        base_tags = dict(
            pipeline_run_id=self.pipeline_run.id,
            pipeline_schedule_id=self.pipeline_run.pipeline_schedule_id,
            pipeline_uuid=self.pipeline.uuid,
        )
        if HOSTNAME:
            base_tags['hostname'] = HOSTNAME
        return merge_dict(kwargs, base_tags)

    def __update_block_run_statuses(self, block_runs: List[BlockRun]) -> None:
        """Update the statuses of the block runs to CONDITION_FAILED or UPSTREAM_FAILED.

        This method updates the statuses of the block runs based on the pipeline run's block runs.
        It retrieves the block UUIDs for failed block runs and conditionally failed block runs.
        It maps the block run statuses to their corresponding block UUIDs.

        The method iterates overthe provided block runs and checks if their dynamic upstream block
        UUIDs or upstream block UUIDs match the failed or conditionally failed block UUIDs.
        * If there is a match, the block run's status is updated accordingly.
        * If no updates are made for a block run, it is added to the list of not updated block runs.

        The method refreshes the pipeline run and continues iterating through block runs until no
        more updates can be made.

        Args:
            block_runs (List[BlockRun]): A list of block runs to update.

        Returns:
            None
        """
        failed_block_uuids = set(
            b.block_uuid for b in self.pipeline_run.block_runs
            if b.status in [
                BlockRun.BlockRunStatus.UPSTREAM_FAILED,
                BlockRun.BlockRunStatus.FAILED,
            ]
        )
        condition_failed_block_uuids = set(
            b.block_uuid for b in self.pipeline_run.block_runs
            if b.status in [
                BlockRun.BlockRunStatus.CONDITION_FAILED,
            ]
        )

        statuses = {
            BlockRun.BlockRunStatus.CONDITION_FAILED: condition_failed_block_uuids,
            BlockRun.BlockRunStatus.UPSTREAM_FAILED: failed_block_uuids,
        }
        not_updated_block_runs = []
        for block_run in block_runs:
            updated_status = False
            dynamic_upstream_block_uuids = block_run.metrics and block_run.metrics.get(
                'dynamic_upstream_block_uuids',
            )

            for status, block_uuids in statuses.items():
                if dynamic_upstream_block_uuids:
                    if all(
                        b in block_uuids
                        for b in dynamic_upstream_block_uuids
                    ):
                        block_run.update(status=status)
                        updated_status = True
                else:
                    block = self.pipeline.get_block(block_run.block_uuid)
                    if any(
                        b in block_uuids
                        for b in block.upstream_block_uuids
                    ):
                        block_run.update(status=status)
                        updated_status = True

            if not updated_status:
                not_updated_block_runs.append(block_run)

        self.pipeline_run.refresh()
        # keep iterating through block runs until no more updates can be made
        if len(block_runs) != len(not_updated_block_runs):
            self.__update_block_run_statuses(not_updated_block_runs)

    def __schedule_blocks(self, block_runs: List[BlockRun] = None) -> None:
        """Schedule the block runs for execution.

        This method schedules the block runs for execution by adding jobs to the job manager.
        It updates the statuses of the initial block runs and fetches any crashed block runs.
        The block runs to be scheduled are determined based on the provided block runs or the
        executable block runs of the pipeline run. The method adds jobs to the job manager for
        each block run, invoking the `run_block` function with the pipeline run ID, block run ID,
        variables, and tags as arguments.

        Args:
            block_runs (List[BlockRun], optional): A list of block runs. Defaults to None.

        Returns:
            None
        """
        self.__update_block_run_statuses(self.pipeline_run.initial_block_runs)
        block_runs_to_schedule = \
            self.executable_block_runs if block_runs is None else block_runs
        block_runs_to_schedule = \
            self.__fetch_crashed_block_runs() + block_runs_to_schedule

        block_run_quota = len(block_runs_to_schedule)
        if self.concurrency_config.block_run_limit is not None:
            queued_or_running_block_runs = self.pipeline_run.queued_or_running_block_runs
            block_run_quota = self.concurrency_config.block_run_limit -\
                len(queued_or_running_block_runs)
            if block_run_quota <= 0:
                return

        for b in block_runs_to_schedule[:block_run_quota]:
            tags = dict(
                block_run_id=b.id,
                block_uuid=b.block_uuid,
            )

            b.update(
                status=BlockRun.BlockRunStatus.QUEUED,
            )

            job_manager.add_job(
                JobType.BLOCK_RUN,
                b.id,
                run_block,
                # args
                self.pipeline_run.id,
                b.id,
                self.pipeline_run.get_variables(),
                self.build_tags(**tags),
            )

    def __schedule_integration_streams(self, block_runs: List[BlockRun] = None) -> None:
        """Schedule the integration streams for execution.

        This method schedules the integration streams for execution by adding jobs to the job
        manager. It determines the integration streams that need to be scheduled based on the
        provided block runs or the pipeline run's block runs. It filters the parallel and
        sequential streams to ensure only streams without corresponding integration stream jobs
        are scheduled. The method generates the necessary variables and runtime arguments for the
        pipeline execution. Jobs are added to the job manager to invoke the `run_integration_stream`
        function for parallel streams and the `run_integration_streams` function for sequential
        streams.

        Args:
            block_runs (List[BlockRun], optional): A list of block runs. Defaults to None.

        Returns:
            None
        """
        if block_runs is not None:
            block_runs_to_schedule = block_runs
        else:
            # Fetch all "in progress" blocks to handle crashed block runs
            block_runs_to_schedule = [
                b for b in self.pipeline_run.block_runs
                if b.status in [
                    BlockRun.BlockRunStatus.INITIAL,
                    BlockRun.BlockRunStatus.QUEUED,
                    BlockRun.BlockRunStatus.RUNNING,
                ]
            ]

        if len(block_runs_to_schedule) > 0:
            tags = self.build_tags()

            block_run_stream_ids = set()
            for br in block_runs_to_schedule:
                stream_id = br.block_uuid.split(':')[-2]
                if stream_id:
                    block_run_stream_ids.add(stream_id)

            filtered_streams = \
                [s for s in self.streams if s['tap_stream_id'] in block_run_stream_ids]
            parallel_streams = list(filter(lambda s: s.get('run_in_parallel'), filtered_streams))
            sequential_streams = list(filter(
                lambda s: not s.get('run_in_parallel'),
                filtered_streams,
            ))

            # Filter parallel streams so that we are only left with block runs for streams
            # that do not have a corresponding integration stream job.
            parallel_streams_to_schedule = []
            for stream in parallel_streams:
                tap_stream_id = stream.get('tap_stream_id')
                if not job_manager.has_integration_stream_job(self.pipeline_run.id, tap_stream_id):
                    parallel_streams_to_schedule.append(stream)

            # Stop scheduling if there are no streams to schedule.
            if (not sequential_streams or job_manager.has_pipeline_run_job(self.pipeline_run.id)) \
                    and len(parallel_streams_to_schedule) == 0:
                return

            # Generate global variables and runtime arguments for pipeline execution.
            variables = self.pipeline_run.get_variables(
                extra_variables=get_extra_variables(self.pipeline),
            )

            pipeline_schedule = self.pipeline_run.pipeline_schedule
            schedule_interval = pipeline_schedule.schedule_interval
            if ScheduleType.API == pipeline_schedule.schedule_type:
                execution_date = datetime.utcnow()
            else:
                # This will be none if trigger is API type
                execution_date = pipeline_schedule.current_execution_date()

            end_date = None
            start_date = None
            date_diff = None

            if ScheduleInterval.ONCE == schedule_interval:
                end_date = variables.get('_end_date')
                start_date = variables.get('_start_date')
            elif ScheduleInterval.HOURLY == schedule_interval:
                date_diff = timedelta(hours=1)
            elif ScheduleInterval.DAILY == schedule_interval:
                date_diff = timedelta(days=1)
            elif ScheduleInterval.WEEKLY == schedule_interval:
                date_diff = timedelta(weeks=1)
            elif ScheduleInterval.MONTHLY == schedule_interval:
                date_diff = relativedelta(months=1)

            if date_diff is not None:
                end_date = (execution_date).isoformat()
                start_date = (execution_date - date_diff).isoformat()

            runtime_arguments = dict(
                _end_date=end_date,
                _execution_date=execution_date.isoformat(),
                _execution_partition=self.pipeline_run.execution_partition,
                _start_date=start_date,
            )

            data_loader_block = self.pipeline.data_loader
            data_exporter_block = self.pipeline.data_exporter
            executable_block_runs = [b.id for b in block_runs_to_schedule]

            self.logger.info(
                f'Start executing PipelineRun {self.pipeline_run.id}: '
                f'pipeline {self.pipeline.uuid}',
                **tags,
            )

            for stream in parallel_streams_to_schedule:
                tap_stream_id = stream.get('tap_stream_id')
                job_manager.add_job(
                    JobType.INTEGRATION_STREAM,
                    f'{self.pipeline_run.id}_{tap_stream_id}',
                    run_integration_stream,
                    # args
                    stream,
                    set(executable_block_runs),
                    tags,
                    runtime_arguments,
                    data_loader_block,
                    data_exporter_block,
                    self.pipeline_run.id,
                    variables,
                )

            if job_manager.has_pipeline_run_job(self.pipeline_run.id) or \
                    len(sequential_streams) == 0:
                return

            job_manager.add_job(
                JobType.PIPELINE_RUN,
                self.pipeline_run.id,
                run_integration_streams,
                # args
                sequential_streams,
                set(executable_block_runs),
                tags,
                runtime_arguments,
                data_loader_block,
                data_exporter_block,
                self.pipeline_run.id,
                variables,
            )

    def __schedule_pipeline(self) -> None:
        """Schedule the pipeline run for execution.

        This method schedules the pipeline run for execution by adding a job to the job manager.
        If a job for the pipeline run already exists, the method returns without scheduling a new
        job. The job added to the job manager invokes the `run_pipeline` function with the
        pipeline run ID, variables, and tags as arguments.

        Returns:
            None
        """
        if job_manager.has_pipeline_run_job(self.pipeline_run.id):
            return
        self.logger.info(
            f'Start a process for PipelineRun {self.pipeline_run.id}',
            **self.build_tags(),
        )
        job_manager.add_job(
            JobType.PIPELINE_RUN,
            self.pipeline_run.id,
            run_pipeline,
            # args
            self.pipeline_run.id,
            self.pipeline_run.get_variables(),
            self.build_tags(),
        )

    def __fetch_crashed_block_runs(self) -> None:
        """Fetch and handle crashed block runs.

        This method fetches the running or queued block runs of the pipeline run and checks if
        their corresponding job is still active. If a job is no longer active, the status of the
        block run is updated to 'INITIAL' to indicate that it needs to be re-executed. A list of
        crashed block runs is returned.

        Returns:
            List[BlockRun]: A list of crashed block runs.
        """
        running_or_queued_block_runs = [b for b in self.pipeline_run.block_runs if b.status in [
            BlockRun.BlockRunStatus.RUNNING,
            BlockRun.BlockRunStatus.QUEUED,
        ]]

        crashed_runs = []
        for br in running_or_queued_block_runs:
            if not job_manager.has_block_run_job(br.id):
                br.update(status=BlockRun.BlockRunStatus.INITIAL)
                crashed_runs.append(br)

        return crashed_runs

    def __run_heartbeat(self) -> None:
        load1, load5, load15, cpu_count = get_compute()
        cpu_usage = load15 / cpu_count if cpu_count else None

        free_memory, used_memory, total_memory = get_memory()
        memory_usage = used_memory / total_memory if total_memory else None

        tags = self.build_tags(
            cpu=load15,
            cpu_total=cpu_count,
            cpu_usage=cpu_usage,
            memory=used_memory,
            memory_total=total_memory,
            memory_usage=memory_usage,
        )

        self.logger.info(
            f'Pipeline {self.pipeline.uuid} for run {self.pipeline_run.id} '
            f'in schedule {self.pipeline_run.pipeline_schedule_id} is alive.',
            **tags,
        )

        if memory_usage and memory_usage >= MEMORY_USAGE_MAXIMUM:
            self.memory_usage_failure(tags)


def run_integration_streams(
    streams: List[Dict],
    *args,
):
    for stream in streams:
        run_integration_stream(stream, *args)


def run_integration_stream(
    stream: Dict,
    executable_block_runs: Set[int],
    tags: Dict,
    runtime_arguments: Dict,
    data_loader_block: Block,
    data_exporter_block: Block,
    pipeline_run_id: int,
    variables: Dict,
):
    """Run an integration stream within the pipeline.

    This method executes an integration stream within the pipeline run. It iterates through each
    stream and executes the corresponding block runs in order. It handles the configuration
    and execution of the data loader, transformer blocks, and data exporter. Metrics calculation is
    performed for the stream if applicable.

    Args:
        stream (Dict): The configuration of the integration stream.
        executable_block_runs (Set[int]): A set of executable block run IDs.
        tags (Dict): A dictionary of tags for logging.
        runtime_arguments (Dict): A dictionary of runtime arguments.
        data_loader_block (Block): The data loader block.
        data_exporter_block (Block): The data exporter block.
        pipeline_run_id (int): The ID of the pipeline run.
        variables (Dict): A dictionary of variables.
    """
    pipeline_run = PipelineRun.query.get(pipeline_run_id)
    pipeline_scheduler = PipelineScheduler(pipeline_run)
    tap_stream_id = stream['tap_stream_id']
    destination_table = stream.get('destination_table', tap_stream_id)

    # all_block_runs is a list of all block runs for the pipeline run
    all_block_runs = BlockRun.query.filter(BlockRun.pipeline_run_id == pipeline_run.id)
    # block_runs is a list of all executable blocks runs for the pipeline run
    block_runs = list(filter(lambda br: br.id in executable_block_runs, all_block_runs))

    # block_runs_for_stream is a list of block runs for the specified stream
    block_runs_for_stream = list(filter(lambda br: tap_stream_id in br.block_uuid, block_runs))
    if len(block_runs_for_stream) == 0:
        return

    indexes = [0]
    for br in block_runs_for_stream:
        parts = br.block_uuid.split(':')
        if len(parts) >= 3:
            indexes.append(int(parts[2]))
    max_index = max(indexes)

    all_block_runs_for_stream = list(filter(
        lambda br: tap_stream_id in br.block_uuid,
        all_block_runs,
    ))
    all_indexes = [0]
    for br in all_block_runs_for_stream:
        # Block run block uuid foramt: "{block_uuid}:{stream_name}:{index}"
        parts = br.block_uuid.split(':')
        if len(parts) >= 3:
            all_indexes.append(int(parts[2]))
    max_index_for_stream = max(all_indexes)

    # Streams can be split up into multiple parts if the source has a large amount of
    # data. Loop through each part of the stream, and execute the blocks runs.
    for idx in range(max_index + 1):
        block_runs_in_order = []
        current_block = data_loader_block

        while True:
            block_runs_in_order.append(
                find(
                    lambda b: b.block_uuid ==
                    f'{current_block.uuid}:{tap_stream_id}:{idx}',  # noqa: B023
                    all_block_runs,
                )
            )
            downstream_blocks = current_block.downstream_blocks
            if len(downstream_blocks) == 0:
                break
            current_block = downstream_blocks[0]

        data_loader_uuid = f'{data_loader_block.uuid}:{tap_stream_id}:{idx}'
        data_exporter_uuid = f'{data_exporter_block.uuid}:{tap_stream_id}:{idx}'

        data_loader_block_run = find(
            lambda b: b.block_uuid == data_loader_uuid,
            all_block_runs,
        )
        data_exporter_block_run = find(
            lambda b: b.block_uuid == data_exporter_uuid,
            block_runs_for_stream,
        )
        if not data_loader_block_run or not data_exporter_block_run:
            continue

        transformer_block_runs = [br for br in block_runs_in_order if (
            br.block_uuid not in [data_loader_uuid, data_exporter_uuid] and
            br.id in executable_block_runs
        )]

        index = stream.get('index', idx)

        # Create config for the block runs. This metadata will be passed into the
        # block before block execution.
        shared_dict = dict(
            destination_table=destination_table,
            index=index,
            is_last_block_run=(index == max_index_for_stream),
            selected_streams=[
                tap_stream_id,
            ],
        )
        block_runs_and_configs = [
            (data_loader_block_run, shared_dict),
        ] + [(br, shared_dict) for br in transformer_block_runs] + [
            (data_exporter_block_run, shared_dict),
        ]
        if len(executable_block_runs) == 1 and \
                data_exporter_block_run.id in executable_block_runs:
            block_runs_and_configs = block_runs_and_configs[-1:]
        elif data_loader_block_run.id not in executable_block_runs:
            block_runs_and_configs = block_runs_and_configs[1:]

        block_failed = False
        for _, tup in enumerate(block_runs_and_configs):
            block_run, template_runtime_configuration = tup

            tags_updated = merge_dict(tags, dict(
                block_run_id=block_run.id,
                block_uuid=block_run.block_uuid,
            ))

            if block_failed:
                block_run.update(
                    status=BlockRun.BlockRunStatus.UPSTREAM_FAILED,
                )
                continue

            pipeline_run.refresh()
            if pipeline_run.status != PipelineRun.PipelineRunStatus.RUNNING:
                return

            block_run.update(
                started_at=datetime.now(),
                status=BlockRun.BlockRunStatus.RUNNING,
            )
            pipeline_scheduler.logger.info(
                f'Start a process for BlockRun {block_run.id}',
                **tags_updated,
            )

            try:
                run_block(
                    pipeline_run_id,
                    block_run.id,
                    variables,
                    tags_updated,
                    pipeline_type=PipelineType.INTEGRATION,
                    verify_output=False,
                    # Not retry for data integration pipeline blocks
                    retry_config=dict(retries=0),
                    runtime_arguments=runtime_arguments,
                    schedule_after_complete=False,
                    template_runtime_configuration=template_runtime_configuration,
                )
            except Exception as e:
                if pipeline_scheduler.allow_blocks_to_fail:
                    block_failed = True
                else:
                    raise e
            else:
                if f'{data_loader_block.uuid}:{tap_stream_id}' in block_run.block_uuid or \
                        f'{data_exporter_block.uuid}:{tap_stream_id}' in block_run.block_uuid:

                    tags2 = merge_dict(tags_updated.get('tags', {}), dict(
                        destination_table=destination_table,
                        index=index,
                        stream=tap_stream_id,
                    ))
                    calculate_metrics(
                        pipeline_run,
                        logger=pipeline_scheduler.logger,
                        logging_tags=merge_dict(tags_updated, dict(tags=tags2)),
                    )


def run_block(
    pipeline_run_id: int,
    block_run_id: int,
    variables: Dict,
    tags: Dict,
    input_from_output: Dict = None,
    pipeline_type: PipelineType = None,
    verify_output: bool = True,
    retry_config: Dict = None,
    runtime_arguments: Dict = None,
    schedule_after_complete: bool = False,
    template_runtime_configuration: Dict = None,
) -> Any:
    """Execute a block within a pipeline run.
    Only run block that's with INITIAL or QUEUED status.

    Args:
        pipeline_run_id (int): The ID of the pipeline run.
        block_run_id (int): The ID of the block run.
        variables (Dict): A dictionary of variables.
        tags (Dict): A dictionary of tags for logging.
        input_from_output (Dict, optional): A dictionary mapping input names to output names.
        pipeline_type (PipelineType, optional): The type of pipeline.
        verify_output (bool, optional): Flag indicating whether to verify the output.
        retry_config (Dict, optional): A dictionary containing retry configuration.
        runtime_arguments (Dict, optional): A dictionary of runtime arguments.
        schedule_after_complete (bool, optional): Flag indicating whether to schedule after
            completion.
        template_runtime_configuration (Dict, optional): A dictionary of template runtime
            configuration.

    Returns:
        Any: The result of executing the block.
    """

    pipeline_run = PipelineRun.query.get(pipeline_run_id)
    if pipeline_run.status != PipelineRun.PipelineRunStatus.RUNNING:
        return {}

    block_run = BlockRun.query.get(block_run_id)
    if block_run.status not in [
        BlockRun.BlockRunStatus.INITIAL,
        BlockRun.BlockRunStatus.QUEUED,
        BlockRun.BlockRunStatus.RUNNING,
    ]:
        return {}

    block_run.update(
        started_at=datetime.now(),
        status=BlockRun.BlockRunStatus.RUNNING,
    )

    pipeline_scheduler = PipelineScheduler(pipeline_run)
    pipeline = pipeline_scheduler.pipeline

    pipeline_scheduler.logger.info(
        f'Execute PipelineRun {pipeline_run.id}, BlockRun {block_run.id}: '
        f'pipeline {pipeline.uuid} block {block_run.block_uuid}',
        **tags)

    if schedule_after_complete:
        on_complete = pipeline_scheduler.on_block_complete
    else:
        on_complete = pipeline_scheduler.on_block_complete_without_schedule

    execution_partition = pipeline_run.execution_partition
    block_uuid = block_run.block_uuid
    block = pipeline.get_block(block_uuid)

    if retry_config is None:
        retry_config = merge_dict(
            get_repo_config(get_repo_path()).retry_config or dict(),
            block.retry_config or dict(),
        )

    return ExecutorFactory.get_block_executor(
        pipeline,
        block_uuid,
        execution_partition=execution_partition,
    ).execute(
        block_run_id=block_run.id,
        global_vars=variables,
        input_from_output=input_from_output,
        on_complete=on_complete,
        on_failure=pipeline_scheduler.on_block_failure,
        pipeline_run_id=pipeline_run_id,
        retry_config=retry_config,
        runtime_arguments=runtime_arguments,
        tags=tags,
        template_runtime_configuration=template_runtime_configuration,
        verify_output=verify_output,
    )


def run_pipeline(pipeline_run_id, variables, tags):
    pipeline_run = PipelineRun.query.get(pipeline_run_id)
    pipeline_scheduler = PipelineScheduler(pipeline_run)
    pipeline = pipeline_scheduler.pipeline
    pipeline_scheduler.logger.info(f'Execute PipelineRun {pipeline_run.id}: '
                                   f'pipeline {pipeline.uuid}',
                                   **tags)
    executor_type = ExecutorFactory.get_pipeline_executor_type(pipeline)
    try:
        pipeline_run.update(executor_type=executor_type)
    except Exception:
        traceback.print_exc()
    ExecutorFactory.get_pipeline_executor(
        pipeline,
        execution_partition=pipeline_run.execution_partition,
        executor_type=executor_type,
    ).execute(
        global_vars=variables,
        pipeline_run_id=pipeline_run_id,
        tags=tags,
    )


def configure_pipeline_run_payload(
    pipeline_schedule: PipelineSchedule,
    pipeline_type: PipelineType,
    payload: Dict = None,
) -> Tuple[Dict, bool]:
    if payload is None:
        payload = dict()
    if 'variables' not in payload:
        payload['variables'] = {}

    payload['pipeline_schedule_id'] = pipeline_schedule.id
    payload['pipeline_uuid'] = pipeline_schedule.pipeline_uuid
    execution_date = payload.get('execution_date')
    if execution_date is None:
        payload['execution_date'] = datetime.utcnow()
    elif not isinstance(execution_date, datetime):
        payload['execution_date'] = datetime.fromisoformat(execution_date)

    # Set execution_partition in variables
    payload['variables']['execution_partition'] = \
        os.sep.join([
            str(pipeline_schedule.id),
            payload['execution_date'].strftime(format='%Y%m%dT%H%M%S_%f'),
        ])

    is_integration = PipelineType.INTEGRATION == pipeline_type
    if is_integration:
        payload['create_block_runs'] = False

    return payload, is_integration


@safe_db_query
def retry_pipeline_run(
    pipeline_run: Dict,
) -> 'PipelineRun':
    pipeline_uuid = pipeline_run['pipeline_uuid']
    pipeline = Pipeline.get(pipeline_uuid, check_if_exists=True)
    if pipeline is None or not pipeline.is_valid_pipeline(pipeline.dir_path):
        raise Exception(f'Pipeline {pipeline_uuid} is not a valid pipeline.')

    pipeline_schedule_id = pipeline_run['pipeline_schedule_id']
    pipeline_run_model = PipelineRun(
        id=pipeline_run['id'],
        pipeline_schedule_id=pipeline_schedule_id,
        pipeline_uuid=pipeline_uuid,
    )
    execution_date = datetime.fromisoformat(pipeline_run['execution_date'])
    new_pipeline_run = pipeline_run_model.create(
        backfill_id=pipeline_run.get('backfill_id'),
        create_block_runs=False,
        execution_date=execution_date,
        event_variables=pipeline_run.get('event_variables', {}),
        pipeline_schedule_id=pipeline_schedule_id,
        pipeline_uuid=pipeline_run_model.pipeline_uuid,
        variables=pipeline_run.get('variables', {}),
    )
    return new_pipeline_run


def stop_pipeline_run(
    pipeline_run: PipelineRun,
    pipeline: Pipeline = None,
) -> None:
    """Stop a pipeline run.

    This function stops a pipeline run by cancelling the pipeline run and its
    associated block runs. If a pipeline object is provided, it also kills the jobs
    associated with the pipeline run and its integration streams if applicable.

    Args:
        pipeline_run (PipelineRun): The pipeline run to stop.
        pipeline (Pipeline, optional): The pipeline associated with the pipeline run.
            Defaults to None.

    Returns:
        None
    """
    if pipeline_run.status not in [PipelineRun.PipelineRunStatus.INITIAL,
                                   PipelineRun.PipelineRunStatus.RUNNING]:
        return

    # Update pipeline run status to cancelled
    pipeline_run.update(status=PipelineRun.PipelineRunStatus.CANCELLED)

    # Cancel all the block runs
    cancel_block_runs_and_jobs(pipeline_run, pipeline)


def cancel_block_runs_and_jobs(
    pipeline_run: PipelineRun,
    pipeline: Pipeline = None,
) -> None:
    """Cancel in progress block runs and jobs for a pipeline run.

    This function cancels blocks runs for the pipeline run. If a pipeline object
    is provided, it also kills the jobs associated with the pipeline run and its
    integration streams if applicable.

    Args:
        pipeline_run (PipelineRun): The pipeline run to stop.
        pipeline (Pipeline, optional): The pipeline associated with the pipeline run.
            Defaults to None.

    Returns:
        None
    """
    block_runs_to_cancel = []
    running_blocks = []
    for b in pipeline_run.block_runs:
        if b.status in [
            BlockRun.BlockRunStatus.INITIAL,
            BlockRun.BlockRunStatus.QUEUED,
            BlockRun.BlockRunStatus.RUNNING,
        ]:
            block_runs_to_cancel.append(b)
        if b.status == BlockRun.BlockRunStatus.RUNNING:
            running_blocks.append(b)
    BlockRun.batch_update_status(
        [b.id for b in block_runs_to_cancel],
        BlockRun.BlockRunStatus.CANCELLED,
    )

    # Kill jobs for integration streams and pipeline run
    if pipeline and pipeline.type in [PipelineType.INTEGRATION, PipelineType.STREAMING]:
        job_manager.kill_pipeline_run_job(pipeline_run.id)
        if pipeline.type == PipelineType.INTEGRATION:
            for stream in pipeline.streams():
                job_manager.kill_integration_stream_job(
                    pipeline_run.id,
                    stream.get('tap_stream_id')
                )
        if pipeline_run.executor_type == ExecutorType.K8S:
            """
            TODO: Support running and cancelling pipeline runs in ECS and GCP_CLOUD_RUN executors
            """
            ExecutorFactory.get_pipeline_executor(
                pipeline,
                executor_type=pipeline_run.executor_type,
            ).cancel(pipeline_run_id=pipeline_run.id)
    else:
        for b in running_blocks:
            job_manager.kill_block_run_job(b.id)


def check_sla():
    repo_pipelines = set(Pipeline.get_all_pipelines(get_repo_path()))
    pipeline_schedules = \
        set([
            s.id
            for s in PipelineSchedule.active_schedules(pipeline_uuids=repo_pipelines)
        ])

    pipeline_runs = PipelineRun.in_progress_runs(pipeline_schedules)

    if pipeline_runs:
        current_time = datetime.now(tz=pytz.UTC)
        # TODO: combine all SLA alerts in one notification
        for pipeline_run in pipeline_runs:
            sla = pipeline_run.pipeline_schedule.sla
            if not sla:
                continue
            start_date = \
                pipeline_run.execution_date \
                if pipeline_run.execution_date is not None \
                else pipeline_run.created_at
            if compare(start_date + timedelta(seconds=sla), current_time) == -1:
                # passed SLA for pipeline_run
                pipeline = Pipeline.get(pipeline_run.pipeline_schedule.pipeline_uuid)
                notification_sender = NotificationSender(
                    NotificationConfig.load(
                        config=merge_dict(
                            pipeline.repo_config.notification_config,
                            pipeline.notification_config,
                        ),
                    ),
                )
                notification_sender.send_pipeline_run_sla_passed_message(
                    pipeline,
                    pipeline_run,
                )

                pipeline_run.update(passed_sla=True)


def schedule_all():
    """
    1. Check whether any new pipeline runs need to be scheduled.
    2. Run git sync if "sync_on_pipeline_run" is enabled.
    3. In active pipeline runs, check whether any block runs need to be scheduled.
    """
    db_connection.session.expire_all()

    repo_pipelines = set(Pipeline.get_all_pipelines(get_repo_path()))

    # Sync schedules from yaml file to DB
    sync_schedules(list(repo_pipelines))

    active_pipeline_schedules = \
        list(PipelineSchedule.active_schedules(pipeline_uuids=repo_pipelines))

    backfills = Backfill.filter(pipeline_schedule_ids=[ps.id for ps in active_pipeline_schedules])

    backfills_by_pipeline_schedule_id = index_by(
        lambda backfill: backfill.pipeline_schedule_id,
        backfills,
    )

    active_pipeline_schedule_ids_with_landing_time_enabled = set()
    for pipeline_schedule in active_pipeline_schedules:
        if pipeline_schedule.landing_time_enabled():
            active_pipeline_schedule_ids_with_landing_time_enabled.add(pipeline_schedule.id)

    previous_pipeline_run_by_pipeline_schedule_id = {}
    if len(active_pipeline_schedule_ids_with_landing_time_enabled) >= 1:
        row_number_column = (
                func.
                row_number().
                over(
                    order_by=desc(PipelineRun.execution_date),
                    partition_by=PipelineRun.pipeline_schedule_id,
                ).
                label('row_number')
        )

        query = PipelineRun.query.filter(
            PipelineRun.pipeline_schedule_id.in_(
                active_pipeline_schedule_ids_with_landing_time_enabled,
            ),
            PipelineRun.status == PipelineRun.PipelineRunStatus.COMPLETED,
        )
        query = query.add_column(row_number_column)
        query = query.from_self().filter(row_number_column == 1)
        for tup in query.all():
            pr, _ = tup
            previous_pipeline_run_by_pipeline_schedule_id[pr.pipeline_schedule_id] = pr

    pipeline_by_uuid = dict()
    concurrency_config_by_uuid = dict()

    git_sync_result = None
    sync_config = get_sync_config()
    for pipeline_schedule in active_pipeline_schedules:
        lock_key = f'pipeline_schedule_{pipeline_schedule.id}'
        if not lock.try_acquire_lock(lock_key):
            continue

        previous_runtimes = []
        if pipeline_schedule.id in active_pipeline_schedule_ids_with_landing_time_enabled:
            previous_pipeline_run = previous_pipeline_run_by_pipeline_schedule_id.get(
                pipeline_schedule.id,
            )
            if previous_pipeline_run:
                previous_runtimes = pipeline_schedule.runtime_history(
                    pipeline_run=previous_pipeline_run,
                )

        # Decide whether to schedule any pipeline runs
        should_schedule = pipeline_schedule.should_schedule(previous_runtimes=previous_runtimes)
        initial_pipeline_runs = [
            r for r in pipeline_schedule.pipeline_runs
            if r.status == PipelineRun.PipelineRunStatus.INITIAL
        ]

        if not should_schedule and not initial_pipeline_runs:
            lock.release_lock(lock_key)
            continue

        pipeline_uuid = pipeline_schedule.pipeline_uuid
        if pipeline_uuid not in pipeline_by_uuid:
            pipeline = pipeline_by_uuid.setdefault(pipeline_uuid, Pipeline.get(pipeline_uuid))
        else:
            pipeline = pipeline_by_uuid[pipeline_uuid]

        if pipeline_uuid not in concurrency_config_by_uuid:
            concurrency_config = concurrency_config_by_uuid.setdefault(
                pipeline_uuid,
                ConcurrencyConfig.load(config=pipeline.concurrency_config),
            )
        else:
            concurrency_config = concurrency_config_by_uuid[pipeline_uuid]

        running_pipeline_runs = [
            r for r in pipeline_schedule.pipeline_runs
            if r.status == PipelineRun.PipelineRunStatus.RUNNING
        ]

        if should_schedule and \
                pipeline_schedule.id not in backfills_by_pipeline_schedule_id:
            # Perform git sync if "sync_on_pipeline_run" is enabled and no other git sync has been
            # run for this scheduler loop.
            if not git_sync_result and sync_config and sync_config.sync_on_pipeline_run:
                git_sync_result = run_git_sync(lock=lock, sync_config=sync_config)

            payload = dict(
                execution_date=pipeline_schedule.current_execution_date(),
                pipeline_schedule_id=pipeline_schedule.id,
                pipeline_uuid=pipeline_uuid,
                variables=pipeline_schedule.variables,
            )

            if len(previous_runtimes) >= 1:
                payload['metrics'] = dict(previous_runtimes=previous_runtimes)

            if PipelineType.INTEGRATION == pipeline.type:
                payload['create_block_runs'] = False

            if pipeline_schedule.get_settings().skip_if_previous_running and \
                    (initial_pipeline_runs or running_pipeline_runs):
                # Cancel the current pipeline run if previous pipeline runs haven't completed
                payload['create_block_runs'] = False
                pipeline_run = PipelineRun.create(**payload)
                pipeline_run.update(status=PipelineRun.PipelineRunStatus.CANCELLED)
            else:
                pipeline_run = PipelineRun.create(**payload)
                # Log Git sync status for new pipeline runs if a git sync result exists
                if git_sync_result:
                    pipeline_scheduler = PipelineScheduler(pipeline_run)
                    log_git_sync(
                        git_sync_result,
                        pipeline_scheduler.logger,
                        pipeline_scheduler.build_tags(),
                    )
                initial_pipeline_runs.append(pipeline_run)

        # Enforce pipeline concurrency limit
        pipeline_run_quota = len(initial_pipeline_runs)
        if concurrency_config.pipeline_run_limit:
            pipeline_run_quota = concurrency_config.pipeline_run_limit - \
                len(running_pipeline_runs)

        if pipeline_run_quota > 0:
            initial_pipeline_runs.sort(key=lambda x: x.execution_date)
            for r in initial_pipeline_runs[:pipeline_run_quota]:
                PipelineScheduler(r).start(should_schedule=False)

        lock.release_lock(lock_key)

    # Schedule active pipeline runs
    active_pipeline_runs = PipelineRun.active_runs_for_pipelines(
        pipeline_uuids=repo_pipelines,
        include_block_runs=True,
    )
    logger.info(f'Active pipeline runs: {[p.id for p in active_pipeline_runs]}')

    for r in active_pipeline_runs:
        try:
            r.refresh()
            PipelineScheduler(r).schedule()
        except Exception:
            logger.exception(f'Failed to schedule {r}')
            traceback.print_exc()
            continue
    job_manager.clean_up_jobs()


def schedule_with_event(event: Dict = None):
    if event is None:
        event = dict()
    logger.info(f'Schedule with event {event}')
    all_event_matchers = EventMatcher.active_event_matchers()

    matched_pipeline_schedules = []
    for e in all_event_matchers:
        if e.match(event):
            logger.info(f'Event matched with {e}')
            matched_pipeline_schedules.extend(e.active_pipeline_schedules())
        else:
            logger.info(f'Event not matched with {e}')

    if len(matched_pipeline_schedules) > 0:
        sync_config = get_sync_config()

        git_sync_result = None
        if sync_config and sync_config.sync_on_pipeline_run:
            git_sync_result = run_git_sync(lock=lock, sync_config=sync_config)

        for p in matched_pipeline_schedules:
            payload = dict(
                execution_date=datetime.now(),
                pipeline_schedule_id=p.id,
                pipeline_uuid=p.pipeline_uuid,
                variables=merge_dict(p.variables or dict(), dict(event=event)),
            )
            pipeline_run = PipelineRun.create(**payload)
            pipeline_scheduler = PipelineScheduler(pipeline_run)

            log_git_sync(
                git_sync_result,
                pipeline_scheduler.logger,
                pipeline_scheduler.build_tags(),
            )

            pipeline_scheduler.start(should_schedule=True)


def sync_schedules(pipeline_uuids: List[str]):
    # Sync schedule configs from triggers.yaml to DB
    for pipeline_uuid in pipeline_uuids:
        pipeline_triggers = get_triggers_by_pipeline(pipeline_uuid)

        logger.debug(f'Sync pipeline trigger configs for {pipeline_uuid}: {pipeline_triggers}.')
        for pipeline_trigger in pipeline_triggers:
            if pipeline_trigger.envs and get_env() not in pipeline_trigger.envs:
                continue
            PipelineSchedule.create_or_update(pipeline_trigger)



from mage_ai.api.resources.GitBranchResource import GitBranchResource
from mage_ai.data_preparation.git import Git


class GitCustomBranchResource(GitBranchResource):
    @classmethod
    def get_git_manager(self, user) -> Git:
        return Git.get_manager(setup_repo=False, user=user)



from mage_ai.api.presenters.GitBranchPresenter import GitBranchPresenter


class GitCustomBranchPresenter(GitBranchPresenter):
    default_attributes = [
        'action_type',
        'error',
        'files',
        'message',
        'modified_files',
        'name',
        'progress',
        'staged_files',
        'status',
        'sync_config',
        'untracked_files',
    ]

    async def present(self, **kwargs):
        display_format = kwargs.get('format')

        data_to_display = self.model

        for key in [
            'modified_files',
            'staged_files',
            'untracked_files',
        ]:
            arr = data_to_display.get(key)
            if arr and type(arr) is list and len(arr) >= 1:
                data_to_display[key] = arr[:100]

        if 'with_basic_details' == display_format:
            return dict(name=data_to_display.get('name'))
        elif 'with_files' == display_format:
            data_to_display.update(
                files=await self.resource.files(
                    self.model.get('modified_files', []),
                    self.model.get('staged_files', []),
                    self.model.get('untracked_files', []),
                    limit=100,
                )
            )
        elif 'with_logs' == display_format:
            data_to_display.update(logs=self.resource.logs(commits=12))
        elif 'with_remotes' == display_format:
            data_to_display.update(remotes=self.resource.remotes(limit=100))

        return data_to_display


GitCustomBranchPresenter.register_format(
    'with_basic_details',
    [
        'name',
    ],
)


GitCustomBranchPresenter.register_format(
    'with_files',
    GitCustomBranchPresenter.default_attributes + [
        'files',
    ],
)


GitCustomBranchPresenter.register_format(
    'with_logs',
    GitCustomBranchPresenter.default_attributes + [
        'logs',
    ],
)


GitCustomBranchPresenter.register_format(
    'with_remotes',
    GitCustomBranchPresenter.default_attributes + [
        'remotes',
    ],
)



from typing import Dict

from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.sync import GitConfig
from mage_ai.data_preparation.sync.git_sync import GitSync
from mage_ai.orchestration.utils.distributed_lock import DistributedLock


def run_git_sync(lock: DistributedLock = None, sync_config: GitConfig = None) -> Dict:
    if sync_config is None:
        preferences = get_preferences()
        if preferences.sync_config:
            sync_config = GitConfig.load(config=preferences.sync_config)

    result = dict()
    git_sync_lock_key = 'git_sync'
    if not lock or lock.try_acquire_lock(git_sync_lock_key):
        try:
            GitSync(sync_config).sync_data()
            result = dict(
                status='success',
                remote_repo_link=sync_config.remote_repo_link,
                branch=sync_config.branch,
            )
        except Exception as err:
            result = dict(
                status='failed',
                error=str(err),
                remote_repo_link=sync_config.remote_repo_link,
                branch=sync_config.branch,
            )
        if lock:
            lock.release_lock(git_sync_lock_key)

    return result


def log_git_sync(result: Dict, logger: DictLogger, tags: Dict = None) -> None:
    if result is None:
        return

    if tags is None:
        tags = dict()
    git_sync_status = result.get('status')
    if git_sync_status == 'success':
        logger.info(
            'Successfully synced data from git repo: '
            f'{result.get("remote_repo_link")}, branch: {result.get("branch")}',
            **tags,
        )
    elif git_sync_status == 'failed':
        logger.warning(
            f'Failed to sync data from git repo: {result.get("remote_repo_link")}'
            f', branch: {result.get("branch")} with error: ' +
            str(result.get("error")),
            **tags,
        )



import os
from typing import Dict, List

from github import Auth, Github

from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.git import Git, api
from mage_ai.data_preparation.preferences import get_preferences


def build_file_object(obj):
    arr = []
    for k, v in obj.items():
        children = build_file_object(v)
        arr.append(dict(
            children=children if len(children) >= 1 else None,
            name=k,
        ))

    return arr


class GitBranchResource(GenericResource):
    @classmethod
    def get_git_manager(self, user) -> Git:
        return Git.get_manager(setup_repo=True, user=user)

    @classmethod
    def collection(self, query, meta, user, **kwargs):
        arr = []

        include_remote_branches = query.get('include_remote_branches', None)
        if include_remote_branches:
            include_remote_branches = include_remote_branches[0]

        repository = query.get('repository', None)
        if repository:
            repository = repository[0]

            access_token = api.get_access_token_for_user(user)
            if access_token:
                auth = Auth.Token(access_token.token)
                g = Github(auth=auth)
                repo = g.get_repo(repository)
                branches = repo.get_branches()

                for branch in branches:
                    arr.append(dict(name=branch.name))
        else:
            git_manager = self.get_git_manager(user=user)
            arr += [dict(name=branch) for branch in git_manager.branches]

            if include_remote_branches:
                pass

        return self.build_result_set(
            arr,
            user,
            **kwargs,
        )

    @classmethod
    def create(self, payload, user, **kwargs):
        branch = payload.get('name')
        git_manager = self.get_git_manager(user=user)
        git_manager.switch_branch(branch)

        return self(dict(name=git_manager.current_branch), user, **kwargs)

    @classmethod
    async def member(self, pk, user, **kwargs):
        git_manager = self.get_git_manager(user=user)

        display_format = kwargs.get('meta', {}).get('_format')
        if 'with_basic_details' == display_format:
            return self(dict(
                files={},
                modified_files=[],
                name=git_manager.current_branch,
                staged_files=[],
                sync_config=get_preferences().sync_config,
                untracked_files=[],
            ), user, **kwargs)

        modified_files = git_manager.modified_files
        staged_files = await git_manager.staged_files()
        untracked_files = await git_manager.untracked_files()

        return self(dict(
            files={},
            modified_files=modified_files,
            name=git_manager.current_branch,
            staged_files=staged_files,
            sync_config=get_preferences().sync_config,
            untracked_files=untracked_files,
        ), user, **kwargs)

    async def update(self, payload, **kwargs):
        git_manager = self.get_git_manager(user=self.current_user)
        action_type = payload.get('action_type')
        files = payload.get('files', None)
        message = payload.get('message', None)
        remote = payload.get('remote', None)

        if action_type == 'status':
            status = git_manager.status()
            untracked_files = await git_manager.untracked_files()
            modified_files = git_manager.modified_files
            self.model = dict(
                name=git_manager.current_branch,
                status=status,
                untracked_files=untracked_files,
                modified_files=modified_files,
            )
        elif action_type == 'commit':
            if not message:
                error = ApiError.RESOURCE_ERROR
                error.update({
                    'message': 'Message is empty, please add a message for your commit.',
                })
                raise ApiError(error)
            git_manager.commit(message, files)
        elif action_type == 'push':
            push = payload.get('push', None)

            if push and 'remote' in push and 'branch' in push:
                from git.exc import GitCommandError

                try:
                    access_token = api.get_access_token_for_user(self.current_user)
                    if access_token:
                        remote = git_manager.repo.remotes[push['remote']]
                        custom_progress = api.push(
                            remote.name,
                            [url for url in remote.urls][0],
                            push['branch'],
                            access_token.token,
                        )
                        if custom_progress.other_lines:
                            lines = custom_progress.other_lines
                            if type(lines) is list:
                                lines = '\n'.join(lines)
                            self.model['progress'] = lines
                    else:
                        self.model['error'] = \
                            'Please authenticate with GitHub before trying to push.'
                except GitCommandError as err:
                    self.model['error'] = str(err)
            else:
                git_manager.push()
        elif action_type == 'pull':
            pull = payload.get('pull', None)
            if pull and 'remote' in pull:
                from git.exc import GitCommandError

                try:
                    access_token = api.get_access_token_for_user(self.current_user)
                    if access_token:
                        branch_name = payload.get('branch')
                        remote = git_manager.repo.remotes[pull['remote']]
                        url = list(remote.urls)[0]

                        custom_progress = None
                        if branch_name:
                            custom_progress = api.pull(
                                remote.name,
                                url,
                                pull.get('branch'),
                                access_token.token,
                            )
                        else:
                            custom_progress = api.fetch(
                                remote.name,
                                url,
                                access_token.token,
                            )

                        if custom_progress and custom_progress.other_lines:
                            lines = custom_progress.other_lines
                            if type(lines) is list:
                                lines = '\n'.join(lines)
                            self.model['progress'] = lines
                    else:
                        self.model['error'] = \
                            'Please authenticate with GitHub before trying to pull.'
                except GitCommandError as err:
                    self.model['error'] = str(err)
            else:
                git_manager.pull()
        elif action_type == 'reset':
            if files and len(files) >= 1:
                for file_path in files:
                    git_manager.reset_file(file_path)
            else:
                git_manager.reset()
        elif action_type == 'clone':
            git_manager.clone()
        elif action_type == 'add':
            for file_path in files:
                git_manager.add_file(file_path, ['-f'])
        elif action_type == 'checkout':
            for file_path in files:
                git_manager.checkout_file(file_path)
        elif action_type in ['add_remote', 'remove_remote']:
            if not remote:
                error = ApiError.RESOURCE_ERROR
                error.update({
                    'message': 'Remote is empty, please add a remote.',
                })
                raise ApiError(error)

            args = []
            keys = ['name']

            if action_type == 'add_remote':
                keys.append('url')

            for key in keys:
                val = remote.get(key)
                if not val:
                    error = ApiError.RESOURCE_ERROR
                    error.update({
                        'message': f'Remote is missing {key}.',
                    })
                    raise ApiError(error)
                args.append(val)

            if action_type == 'add_remote':
                git_manager.add_remote(*args)
            elif action_type == 'remove_remote':
                git_manager.remove_remote(*args)
        elif action_type in [
            'delete',
            'merge',
            'rebase',
        ]:
            data = payload.get('delete', None) or \
                payload.get('merge', None) or \
                payload.get('rebase', None)

            if data and 'base_branch' in data:
                base_branch = data['base_branch']

                if 'delete' == action_type:
                    git_manager.delete_branch(base_branch)
                elif 'merge' == action_type:
                    git_manager.merge_branch(base_branch, message=message)
                elif 'rebase' == action_type:
                    git_manager.rebase_branch(base_branch, message=message)
            else:
                error = ApiError.RESOURCE_ERROR
                error.update({
                    'message': 'Please select a base branch to ' +
                    f'{action_type} into the current branch.',
                })
                raise ApiError(error)

        return self

    async def files(
        self,
        modified_files: List[str],
        staged_files: List[str],
        untracked_files: List[str],
        limit: int = None
    ) -> Dict:
        arr = modified_files + staged_files + untracked_files

        if limit:
            arr = arr[:limit]

        files = {}
        for filename in arr:
            # filename: default_repo/transformers/load.py
            parts = filename.split(os.sep)
            number_of_parts = len(parts)

            arr = []
            for idx, part in enumerate(parts):
                default_obj = dict()

                if idx == 0:
                    obj = files.get(part, default_obj)
                else:
                    obj_prev = arr[idx - 1]
                    obj = obj_prev.get(part, default_obj)

                arr.append(obj)

            obj_final = None
            for idx, obj in enumerate(reversed(arr)):
                if idx == 0:
                    obj_final = obj
                else:
                    part = parts[number_of_parts - idx]
                    obj[part] = obj_final
                    obj_final = obj

            files[parts[0]] = obj_final

        return build_file_object(files)

    def logs(self, commits: int = None) -> List[Dict]:
        git_manager = self.get_git_manager(user=self.current_user)
        return git_manager.logs(commits=commits)

    def remotes(self, limit: int = None) -> List[Dict]:
        git_manager = self.get_git_manager(user=self.current_user)
        return git_manager.remotes(limit=limit, user=self.current_user)



import asyncio

import typer

from mage_ai.ai.llm_pipeline_wizard import LLMPipelineWizard

app = typer.Typer()

PROJECT_NAME_DEFAULT = typer.Argument(
    ..., help='Mage project path'
)
PIPELINE_NAME_DEFAULT = typer.Argument(
    ..., help='Mage pipeline uuid'
)
BLOCK_NAME_DEFAULT = typer.Argument(
    ..., help='Mage block uuid in the pipeline'
)
PRINT_BLOCK_DOC_DEFAULT = typer.Option(
    False, help='specify if block documentation should be printed.'
)


@app.command()
def generate_block_documentation(project_path: str = PROJECT_NAME_DEFAULT,
                                 pipeline_uuid: str = PROJECT_NAME_DEFAULT,
                                 block_uuid: str = BLOCK_NAME_DEFAULT):
    print(asyncio.run(LLMPipelineWizard().async_generate_block_documentation_with_name(
        pipeline_uuid,
        block_uuid,
        project_path=project_path,
    )))


@app.command()
def generate_pipeline_documentation(project_path: str = PROJECT_NAME_DEFAULT,
                                    pipeline_uuid: str = PROJECT_NAME_DEFAULT,
                                    print_block_doc: bool = PRINT_BLOCK_DOC_DEFAULT):
    print(asyncio.run(LLMPipelineWizard().async_generate_pipeline_documentation(
        pipeline_uuid,
        print_block_doc=print_block_doc,
        project_path=project_path,
    ))['pipeline_doc'])


@app.command()
def generate_block_with_description(block_description: str):
    print(asyncio.run(LLMPipelineWizard().async_generate_block_with_description(block_description)))


if __name__ == '__main__':
    app()



import json
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union

from ldap3 import Connection, Server
from ldap3.core.exceptions import LDAPException

from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.settings import (
    LDAP_AUTHENTICATION_FILTER,
    LDAP_AUTHORIZATION_FILTER,
    LDAP_BASE_DN,
    LDAP_BIND_DN,
    LDAP_BIND_PASSWORD,
    LDAP_GROUP_FIELD,
    LDAP_ROLES_MAPPING,
    LDAP_SERVER,
)
from mage_ai.shared.hash import merge_dict


class LDAPAuthenticator(ABC):
    @abstractmethod
    def authenticate(self, username: str, password: str) -> Tuple[bool, str]:
        pass

    @abstractmethod
    def authorize(self, username: str) -> bool:
        pass

    def verify(self, username: str, password: str) -> bool:
        authenticated, user_dn = self.authenticate(username, password)
        if authenticated:
            return self.authorize(user_dn)
        return False


class LDAPConnection(LDAPAuthenticator):
    def __init__(
        self,
        server_url: str,
        bind_dn: str,
        bind_password: str,
        base_dn: str,
        authentication_filter: str,
        authorization_filter: str,
        group_field: str = None,
        roles_mapping: Union[Dict, str] = None,
    ):
        self.server_url = server_url
        self.bind_dn = bind_dn
        self.bind_password = bind_password
        self.base_dn = base_dn
        self.authentication_filter = authentication_filter
        self.authorization_filter = authorization_filter
        self.group_field = group_field

        if type(roles_mapping) is str:
            try:
                self.roles_mapping = json.loads(roles_mapping)
            except json.JSONDecodeError:
                self.roles_mapping = None
        else:
            self.roles_mapping = roles_mapping

        self.conn = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.unbind()

    def bind(self):
        self.conn = Connection(
            Server(self.server_url), self.bind_dn, self.bind_password, auto_bind=True
        )

    def authenticate(self, username: str, password: str) -> Tuple[bool, str, Dict]:
        try:
            if not self.conn:
                self.bind()
            self.conn.search(
                self.base_dn,
                self.authentication_filter.format(username=username),
                attributes=[self.group_field] if self.group_field else None,
            )
            if not self.conn.entries:
                return False, ""
            user_dn = self.conn.entries[0].entry_dn
            user_attributes = self.conn.entries[0].entry_attributes_as_dict
            if self.conn.rebind(user=user_dn, password=password):
                return True, user_dn, user_attributes
            return False, user_dn, user_attributes
        except LDAPException:
            return False, "", dict()

    def authorize(self, user_dn: str) -> bool:
        try:
            if not self.conn:
                self.bind()
            self.conn.entries.clear()
            self.conn.search(self.base_dn, self.authorization_filter.format(user_dn=user_dn))
            if self.conn.entries:
                if len(self.conn.entries) > 0:
                    return True
            return False
        except LDAPException:
            return False

    def get_user_roles(self, user_attributes: Dict[str, List]) -> List[str]:
        if not self.roles_mapping:
            return []
        groups = user_attributes.get(self.group_field, [])
        roles = set()
        for group in groups:
            if group in self.roles_mapping:
                roles.update(self.roles_mapping[group])
            else:
                print(f"Can't find group {group} in LDAP role mapping")
        return list(roles)


def new_ldap_connection() -> LDAPConnection:
    ldap_config_from_env_vars = dict(
        server_url=LDAP_SERVER,
        bind_dn=LDAP_BIND_DN,
        bind_password=LDAP_BIND_PASSWORD,
        base_dn=LDAP_BASE_DN,
        authentication_filter=LDAP_AUTHENTICATION_FILTER,
        authorization_filter=LDAP_AUTHORIZATION_FILTER,
        group_field=LDAP_GROUP_FIELD,
        roles_mapping=LDAP_ROLES_MAPPING,
    )
    try:
        ldap_config = get_repo_config().ldap_config
    except Exception:
        ldap_config = dict()
    merged = merge_dict(ldap_config_from_env_vars, ldap_config)

    return LDAPConnection(**merged)


def verify(username, password: str) -> bool:
    return new_ldap_connection().verify(username, password)



import asyncio
import requests
import subprocess
from datetime import datetime
from git.remote import RemoteProgress
from git.repo.base import Repo
from mage_ai.authentication.oauth.constants import OAUTH_PROVIDER_GITHUB
from mage_ai.orchestration.db.models.oauth import Oauth2AccessToken, Oauth2Application, User
from typing import Dict
from urllib.parse import urlsplit, urlunsplit


API_ENDPOINT = 'https://api.github.com'


def get_access_token_for_user(user: User) -> Oauth2AccessToken:
    oauth_client = Oauth2Application.query.filter(
        Oauth2Application.client_id == OAUTH_PROVIDER_GITHUB,
    ).first()

    if oauth_client:
        access_token = Oauth2AccessToken.query.filter(
            Oauth2AccessToken.expires > datetime.utcnow(),
            Oauth2AccessToken.oauth2_application_id == oauth_client.id,
        ).first()

        return access_token


def fetch(remote_name: str, remote_url: str, token: str) -> RemoteProgress:
    from mage_ai.data_preparation.git import Git

    custom_progress = RemoteProgress()
    username = get_username(token)

    url = build_authenticated_remote_url(remote_url, username, token)
    git_manager = Git.get_manager()

    remote = git_manager.repo.remotes[remote_name]
    url_original = list(remote.urls)[0]
    remote.set_url(url)

    try:
        remote.fetch(progress=custom_progress)
    except Exception as err:
        raise err
    finally:
        try:
            remote.set_url(url_original)
        except Exception as err:
            print('WARNING (mage_ai.data_preparation.git.api):')
            print(err)

    return custom_progress


def pull(remote_name: str, remote_url: str, branch_name: str, token: str) -> RemoteProgress:
    from mage_ai.data_preparation.git import Git

    custom_progress = RemoteProgress()
    username = get_username(token)

    url = build_authenticated_remote_url(remote_url, username, token)
    git_manager = Git.get_manager()

    remote = git_manager.repo.remotes[remote_name]
    url_original = list(remote.urls)[0]
    remote.set_url(url)

    try:
        remote.pull(branch_name, custom_progress)
    except Exception as err:
        raise err
    finally:
        try:
            remote.set_url(url_original)
        except Exception as err:
            print('WARNING (mage_ai.data_preparation.git.api):')
            print(err)

    return custom_progress


def push(remote_name: str, remote_url: str, branch_name: str, token: str) -> RemoteProgress:
    from mage_ai.data_preparation.git import Git

    custom_progress = RemoteProgress()
    username = get_username(token)

    url = build_authenticated_remote_url(remote_url, username, token)
    git_manager = Git.get_manager()

    remote = git_manager.repo.remotes[remote_name]
    url_original = list(remote.urls)[0]
    remote.set_url(url)

    try:
        remote.push(branch_name, custom_progress)
    except Exception as err:
        raise err
    finally:
        try:
            remote.set_url(url_original)
        except Exception as err:
            print('WARNING (mage_ai.data_preparation.git.api):')
            print(err)

    return custom_progress


def build_authenticated_remote_url(remote_url: str, username: str, token: str) -> str:
    # https://[username]:[token]@github.com/[remote_url]
    url = urlsplit(remote_url)
    url = url._replace(netloc=f'{username}:{token}@{url.netloc}')
    return urlunsplit(url)


def get_username(token: str) -> str:
    return get_user(token)['login']


def get_user(token: str) -> Dict:
    """
    https://docs.github.com/en/rest/users/users?apiVersion=2022-11-28#get-the-authenticated-user
    """
    resp = requests.get(f'{API_ENDPOINT}/user', headers={
        'Accept': 'application/vnd.github+json',
        'Authorization': f'Bearer {token}',
        'X-GitHub-Api-Version': '2022-11-28',
    })

    return resp.json()


def check_connection(repo: Repo, remote_url: str) -> None:
    asyncio.run(validate_authentication_for_remote_url(repo, remote_url))


async def validate_authentication_for_remote_url(repo: Repo, remote_url: str) -> None:
    proc = repo.git.ls_remote(remote_url, as_process=True)

    asyncio.run(__poll_process_with_timeout(
        proc,
        error_message='Error connecting to remote, make sure your access is valid.',
    ))


async def __poll_process_with_timeout(
    proc: subprocess.Popen,
    error_message: str = None,
    timeout: int = 10,
):
    ct = 0
    while ct < timeout * 2:
        return_code = proc.poll()
        if return_code is not None:
            proc.kill()
            break
        ct += 1
        await asyncio.sleep(0.5)

    if error_message is None:
        error_message = 'Error running Git process'

    if return_code is not None and return_code != 0:
        _, err = proc.communicate()
        message = (
            err.decode('UTF-8') if err
            else error_message
        )
        raise ChildProcessError(message)

    if return_code is None:
        proc.kill()
        raise TimeoutError(error_message)



import os
import shutil
import yaml
from dataclasses import dataclass, field
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.custom_templates.constants import (
    DIRECTORY_FOR_PIPELINE_TEMPLATES,
    METADATA_FILENAME_WITH_EXTENSION,
)
from mage_ai.data_preparation.models.custom_templates.utils import custom_templates_directory
from mage_ai.data_preparation.models.triggers import (
    TRIGGER_FILE_NAME,
    Trigger,
    add_or_update_trigger_for_pipeline_and_persist,
    get_triggers_by_pipeline,
    load_trigger_configs,
)
from mage_ai.orchestration.db.models.schedules import PipelineSchedule
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.config import BaseConfig
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.io import safe_write
from mage_ai.shared.utils import clean_name
from typing import Dict, List


@dataclass
class CustomPipelineTemplate(BaseConfig):
    description: str = None
    name: str = None
    pipeline: Dict = field(default_factory=dict)
    tags: List = field(default_factory=list)
    template_uuid: str = None
    user: Dict = field(default_factory=dict)

    @classmethod
    def load(self, template_uuid: str = None, uuid: str = None):
        uuid_use = uuid
        template_uuid_use = template_uuid

        if uuid_use:
            parts = uuid_use.split(os.sep)
            template_uuid_use = os.path.join(*parts[2:])
        elif template_uuid_use:
            uuid_use = os.path.join(
                custom_templates_directory(),
                DIRECTORY_FOR_PIPELINE_TEMPLATES,
                template_uuid_use,
            )

        try:
            config_path_metadata = os.path.join(
                get_repo_path(),
                uuid_use,
                METADATA_FILENAME_WITH_EXTENSION,
            )
            custom_template = super().load(config_path_metadata)
            custom_template.template_uuid = template_uuid_use

            return custom_template
        except Exception as err:
            print(f'[WARNING] CustomPipelineTemplate.load: {err}')

    @classmethod
    def create_from_pipeline(
        self,
        pipeline: Pipeline,
        template_uuid: str,
        name: str = None,
        description: str = None,
    ):
        pipeline_dict = pipeline.to_dict(
            exclude_data_integration=True,
            include_extensions=True,
        )

        custom_template = self(
            description=description,
            name=name,
            pipeline=pipeline_dict,
            template_uuid=clean_name(template_uuid, [os.sep]) if template_uuid else template_uuid,
        )

        custom_template.save()

        triggers = get_triggers_by_pipeline(pipeline.uuid)

        pipeline_schedules = PipelineSchedule.query.filter(
            PipelineSchedule.pipeline_uuid == pipeline.uuid,
        ).all()
        for pipeline_schedule in pipeline_schedules:
            trigger = Trigger(
                name=pipeline_schedule.name,
                pipeline_uuid=pipeline_schedule.pipeline_uuid,
                schedule_interval=pipeline_schedule.schedule_interval,
                schedule_type=pipeline_schedule.schedule_type,
                settings=pipeline_schedule.settings,
                sla=pipeline_schedule.sla,
                start_time=pipeline_schedule.start_time,
                status=pipeline_schedule.status,
                variables=pipeline_schedule.variables,
            )
            triggers.append(trigger)

        if triggers:
            custom_template.save_triggers(triggers)

        return custom_template

    @property
    def uuid(self):
        return os.path.join(
            custom_templates_directory(),
            DIRECTORY_FOR_PIPELINE_TEMPLATES,
            self.template_uuid,
        )

    @property
    def metadata_file_path(self) -> str:
        return os.path.join(
            get_repo_path(),
            self.uuid,
            METADATA_FILENAME_WITH_EXTENSION,
        )

    @property
    def triggers_file_path(self) -> str:
        return os.path.join(
            get_repo_path(),
            self.uuid,
            TRIGGER_FILE_NAME,
        )

    def build_pipeline(self) -> Pipeline:
        return Pipeline(clean_name(self.template_uuid), config=self.pipeline)

    def create_pipeline(self, name: str) -> Pipeline:
        pipeline = Pipeline(clean_name(name), config=self.pipeline)
        os.makedirs(os.path.dirname(pipeline.config_path), exist_ok=True)
        pipeline.save()

        if os.path.isfile(self.triggers_file_path):
            pipeline_uuid = pipeline.uuid

            with open(self.triggers_file_path, 'r') as f:
                content = f.read()

                for trigger in load_trigger_configs(content, pipeline_uuid=pipeline_uuid):
                    add_or_update_trigger_for_pipeline_and_persist(trigger, pipeline_uuid)

        return pipeline

    def to_dict(self) -> Dict:
        return merge_dict(self.to_dict_base(), dict(
            template_uuid=self.template_uuid,
            uuid=self.uuid,
        ))

    def to_dict_base(self) -> Dict:
        return dict(
            description=self.description,
            name=self.name,
            pipeline=self.pipeline,
            tags=self.tags,
            user=self.user,
        )

    def save(self) -> None:
        content = yaml.safe_dump(self.to_dict_base())
        file_path = self.metadata_file_path
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        safe_write(file_path, content)

    def save_triggers(self, triggers: List[Dict]) -> None:
        content = yaml.safe_dump(dict(triggers=[trigger.to_dict() for trigger in triggers]))
        file_path = self.triggers_file_path
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        safe_write(file_path, content)

    def delete(self) -> None:
        shutil.rmtree(os.path.join(get_repo_path(), self.uuid))



from mage_ai.api.presenters.BasePresenter import BasePresenter


class PullRequestPresenter(BasePresenter):
    default_attributes = [
        'body',
        'created_at',
        'id',
        'is_merged',
        'last_modified',
        'merged',
        'state',
        'title',
        'url',
        'user',
    ]



ACTIVE_DIRECTORY_CLIENT_ID = '51aec820-9d49-40a9-b046-17c1f28f620d'

GITHUB_CLIENT_ID = '8577f13ddc81e2848b07'
GITHUB_STATE = '1337'

OAUTH_PROVIDER_ACTIVE_DIRECTORY = 'active_directory'
OAUTH_PROVIDER_GITHUB = 'github'

VALID_OAUTH_PROVIDERS = [
    OAUTH_PROVIDER_ACTIVE_DIRECTORY,
    OAUTH_PROVIDER_GITHUB,
]



import enum

DATABASE_CONNECTION_URL_ENV_VAR = 'MAGE_DATABASE_CONNECTION_URL'

DB_USER = 'DB_USER'
DB_PASS = 'DB_PASS'
DB_NAME = 'DB_NAME'

PIPELINE_RUN_MAGE_VARIABLES_KEY = '__mage_variables'


class Entity(str, enum.Enum):
    # Permissions saved to the DB should not have the "ANY" entity. It should only be used
    # when evaluating permissions.
    ANY = 'any'
    GLOBAL = 'global'
    PROJECT = 'project'
    PIPELINE = 'pipeline'



import os
from mage_ai.cache.base import BaseCache
from mage_ai.cache.block_action_object.constants import (
    OBJECT_TYPE_BLOCK_FILE,
    OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE,
    OBJECT_TYPE_MAGE_TEMPLATE,
)
from mage_ai.cache.constants import CACHE_KEY_BLOCK_ACTION_OBJECTS_MAPPING
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import (
    BlockLanguage,
    BlockType,
    FILE_EXTENSION_TO_BLOCK_LANGUAGE,
)
from mage_ai.data_preparation.models.custom_templates.constants import (
    DIRECTORY_FOR_BLOCK_TEMPLATES,
)
from mage_ai.data_preparation.models.custom_templates.custom_block_template import (
    CustomBlockTemplate,
)
from mage_ai.data_preparation.models.custom_templates.utils import (
    flatten_files,
    get_templates,
    group_and_hydrate_files,
)
from mage_ai.data_preparation.templates.constants import (
    TEMPLATES,
    TEMPLATES_ONLY_FOR_V2,
)
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.strings import remove_extension_from_filename
from pathlib import Path
from typing import Dict


def parse_block_file_absolute_path(block_file_absolute_path: str) -> Dict:
    block_type = Block.block_type_from_path(block_file_absolute_path)
    file_directory_name = Block.file_directory_name(block_type)

    # This is the block_file_absolute_path without the repo_path
    file_path = str(block_file_absolute_path).replace(get_repo_path(), '')
    if file_path.startswith(os.sep):
        file_path = file_path[1:]

    file_path_parts = file_path.split(os.path.sep)

    if len(file_path_parts) >= 1 and file_directory_name == file_path_parts[0]:
        file_path_parts = file_path_parts[1:]

    # This is the file_path without the block type directory
    filename = os.path.join(*file_path_parts)

    filename_parts = filename.split('.')
    file_extension = None
    if len(filename_parts) >= 2:
        file_extension = filename_parts[-1]

    block_language = None
    if file_extension:
        block_language = FILE_EXTENSION_TO_BLOCK_LANGUAGE.get(file_extension)

    return dict(
        block_type=block_type,
        file_path=file_path,
        filename=filename,
        filename_parts=filename_parts,
        language=block_language,
    )


class BlockActionObjectCache(BaseCache):
    cache_key = CACHE_KEY_BLOCK_ACTION_OBJECTS_MAPPING

    @classmethod
    async def initialize_cache(self, replace: bool = False) -> 'BlockActionObjectCache':
        cache = self()
        if replace or not cache.exists():
            await cache.initialize_cache_for_all_objects()

        return cache

    def build_key_for_block_file(
        self,
        file_path: str = None,
        block_type: BlockType = None,
        language: BlockLanguage = None,
        filename: str = None,
    ) -> str:
        return ':'.join(list(filter(lambda x: x, [
            file_path,
            block_type,
            language,
            remove_extension_from_filename(filename).replace('_', ' ') if filename else filename,
        ])))

    def build_key_for_custom_block_template(
        self,
        custom_block_template: CustomBlockTemplate,
    ) -> str:
        arr = []

        # This is a specific order
        for key in [
            'uuid',
            'block_type',
            'language',
            'name',
            'description',
            'tags',
        ]:
            val = getattr(custom_block_template, key)
            if val:
                if not isinstance(val, list):
                    val = [val]
                arr += val

        return ':'.join(arr)

    def build_key_for_mage_template(self, block_action_object: Dict) -> str:
        arr = []

        # This is a specific order
        for key in [
            'path',
            'block_type',
            'language',
            'name',
            'description',
            'groups',
        ]:
            val = block_action_object.get(key)
            if val:
                if not isinstance(val, list):
                    val = [val]
                arr += val

        return ':'.join(arr)

    def update_block(
        self,
        block: Block = None,
        block_file_absolute_path: str = None,
        remove: bool = False,
    ) -> None:
        if block:
            block_type = block.type
            file_path = os.path.join(block.file.dir_path, block.file.filename)
            filename = os.path.join(*file_path.split(os.path.sep)[1:])
            language = block.language
            uuid = block.uuid
        else:
            d = parse_block_file_absolute_path(block_file_absolute_path)
            block_type = d.get('block_type')
            file_path = d.get('file_path')
            filename = d.get('filename')
            language = d.get('language')

            filename_parts = d.get('filename_parts')
            uuid = '.'.join(filename_parts[:-1])

        key = self.build_key_for_block_file(
            file_path,
            block_type,
            language,
            filename,
        )

        mapping = self.load_all_data()
        if mapping is None:
            mapping = {}

        if OBJECT_TYPE_BLOCK_FILE not in mapping:
            mapping[OBJECT_TYPE_BLOCK_FILE] = {}

        if remove:
            if key in mapping[OBJECT_TYPE_BLOCK_FILE]:
                mapping[OBJECT_TYPE_BLOCK_FILE].pop(key, None)
        else:
            content = None
            if block:
                content = block.content
            else:
                with open(block_file_absolute_path, 'r') as f:
                    content = f.read()

            mapping[OBJECT_TYPE_BLOCK_FILE][key] = dict(
                content=content,
                file_path=file_path,
                language=language,
                type=block_type,
                uuid=uuid,
            )

        self.set(self.cache_key, mapping)

    def update_custom_block_template(
        self,
        custom_block_template: CustomBlockTemplate,
        remove: bool = False,
    ) -> None:
        key = self.build_key_for_custom_block_template(custom_block_template)

        mapping = self.load_all_data()
        if mapping is None:
            mapping = {}

        if OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE not in mapping:
            mapping[OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE] = {}

        if remove:
            if key in mapping[OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE]:
                mapping[OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE].pop(key, None)
        else:
            mapping[OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE][key] = custom_block_template.to_dict(
                include_content=True,
            )

        self.set(self.cache_key, mapping)

    async def initialize_cache_for_all_objects(self) -> None:
        mapping = {
            OBJECT_TYPE_BLOCK_FILE: {},
            OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE: {},
            OBJECT_TYPE_MAGE_TEMPLATE: {},
        }

        for block_action_object in (TEMPLATES + TEMPLATES_ONLY_FOR_V2):
            key = self.build_key_for_mage_template(block_action_object)
            mapping[OBJECT_TYPE_MAGE_TEMPLATE][key] = block_action_object

        file_dicts = get_templates(DIRECTORY_FOR_BLOCK_TEMPLATES)
        file_dicts_flat = flatten_files(file_dicts)
        custom_block_templates = group_and_hydrate_files(file_dicts_flat, CustomBlockTemplate)

        for custom_block_template in custom_block_templates:
            key = self.build_key_for_custom_block_template(custom_block_template)
            mapping[OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE][key] = custom_block_template.to_dict(
                include_content=True,
            )

        for block_type in BlockType:
            file_directory_name = Block.file_directory_name(block_type)
            directory_full_path = os.path.join(get_repo_path(), file_directory_name)

            for path in Path(directory_full_path).rglob('*'):
                if not path.is_file():
                    continue

                block_file_absolute_path = path.absolute()
                d = parse_block_file_absolute_path(block_file_absolute_path)

                file_path = d.get('file_path')
                filename = d.get('filename')
                filename_parts = d.get('filename_parts')
                language = d.get('language')

                if '__init__.py' == filename:
                    continue

                if not language:
                    continue

                key = self.build_key_for_block_file(
                    file_path,
                    block_type,
                    language,
                    filename,
                )

                content = None
                with open(block_file_absolute_path, 'r') as f:
                    content = f.read()

                mapping[OBJECT_TYPE_BLOCK_FILE][key] = dict(
                    content=content,
                    file_path=file_path,
                    language=language,
                    type=block_type,
                    uuid='.'.join(filename_parts[:-1]),
                )

        self.set(self.cache_key, mapping)
